from ruamel.yaml import YAML

typ = "rt"
yaml = YAML(typ=typ)
yaml.preserve_quotes = True
yaml.indent(mapping=2, sequence=4, offset=2)
yaml.representer.ignore_aliases = lambda *data: True
yaml.width = 4096



import sys
import time
from collections import defaultdict
from typing import Dict, List, Optional

import sentry_sdk
from pydantic.json import pydantic_encoder

from consoleme.config import config
from consoleme.lib.cache import (
    retrieve_json_data_from_redis_or_s3,
    store_json_results_in_redis_and_s3,
)
from consoleme.lib.cloud_credential_authorization_mapping.dynamic_config import (
    DynamicConfigAuthorizationMappingGenerator,
)
from consoleme.lib.cloud_credential_authorization_mapping.internal_plugin import (
    InternalPluginAuthorizationMappingGenerator,
)
from consoleme.lib.cloud_credential_authorization_mapping.models import (
    RoleAuthorizations,
    RoleAuthorizationsDecoder,
    user_or_group,
)
from consoleme.lib.cloud_credential_authorization_mapping.role_tags import (
    RoleTagAuthorizationMappingGenerator,
)
from consoleme.lib.singleton import Singleton

log = config.get_logger("consoleme")


class CredentialAuthorizationMapping(metaclass=Singleton):
    def __init__(self) -> None:
        self._all_roles = []
        self._all_roles_count = 0
        self._all_roles_last_update = 0
        self.authorization_mapping = {}
        self.authorization_mapping_last_update = 0
        self.reverse_mapping = {}
        self.reverse_mapping_last_update = 0

    async def retrieve_credential_authorization_mapping(
        self, max_age: Optional[int] = None
    ):
        """
        This function retrieves the credential authorization mapping. This is a mapping of users/groups to the IAM roles
        they are allowed to get credentials for. This is the authoritative mapping that ConsoleMe uses for access.

        :param max_age: Maximum allowable age of the credential authorization mapping. If the mapping is older than
        `max_age` seconds, this function will raise an exception and return an empty mapping.
        """
        if (
            not self.authorization_mapping
            or int(time.time()) - self.authorization_mapping_last_update > 60
        ):
            redis_topic = config.get(
                "generate_and_store_credential_authorization_mapping.redis_key",
                "CREDENTIAL_AUTHORIZATION_MAPPING_V1",
            )
            s3_bucket = config.get(
                "generate_and_store_credential_authorization_mapping.s3.bucket"
            )
            s3_key = config.get(
                "generate_and_store_credential_authorization_mapping.s3.file",
                "credential_authorization_mapping/credential_authorization_mapping_v1.json.gz",
            )
            try:
                self.authorization_mapping = await retrieve_json_data_from_redis_or_s3(
                    redis_topic,
                    s3_bucket=s3_bucket,
                    s3_key=s3_key,
                    json_object_hook=RoleAuthorizationsDecoder,
                    json_encoder=pydantic_encoder,
                    max_age=max_age,
                )
                self.authorization_mapping_last_update = int(time.time())
            except Exception as e:
                sentry_sdk.capture_exception()
                log.error(
                    {
                        "function": f"{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
                        "error": f"Error loading cloud credential mapping. Returning empty mapping: {e}",
                    },
                    exc_info=True,
                )
                return {}
        return self.authorization_mapping

    async def retrieve_reverse_authorization_mapping(
        self, max_age: Optional[int] = None
    ):
        """
        This function retrieves the inverse of the credential authorization mapping. This is a mapping of IAM roles
        to the users/groups that are allowed to access them. This mapping is used primarily for auditing.

        :param max_age: Maximum allowable age of the reverse credential authorization mapping. If the mapping is older
        than `max_age` seconds, this function will raise an exception and return an empty mapping.
        """
        if (
            not self.reverse_mapping
            or int(time.time()) - self.reverse_mapping_last_update > 60
        ):
            redis_topic = config.get(
                "generate_and_store_reverse_authorization_mapping.redis_key",
                "REVERSE_AUTHORIZATION_MAPPING_V1",
            )
            s3_bucket = config.get(
                "generate_and_store_reverse_authorization_mapping.s3.bucket"
            )
            s3_key = config.get(
                "generate_and_store_reverse_authorization_mapping.s3.file",
                "reverse_authorization_mapping/reverse_authorization_mapping_v1.json.gz",
            )
            try:
                self.reverse_mapping = await retrieve_json_data_from_redis_or_s3(
                    redis_topic,
                    s3_bucket=s3_bucket,
                    s3_key=s3_key,
                    json_object_hook=RoleAuthorizationsDecoder,
                    json_encoder=pydantic_encoder,
                    max_age=max_age,
                )
                self.reverse_mapping_last_update = int(time.time())
            except Exception as e:
                sentry_sdk.capture_exception()
                log.error(
                    {
                        "function": f"{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
                        "error": f"Error loading reverse credential mapping. Returning empty mapping: {e}",
                    },
                    exc_info=True,
                )
                return {}
        return self.reverse_mapping

    async def retrieve_all_roles(self, max_age: Optional[int] = None):
        if not self._all_roles or int(time.time()) - self._all_roles_last_update > 600:
            redis_topic = config.get("aws.iamroles_redis_key", "IAM_ROLE_CACHE")
            s3_bucket = config.get(
                "cache_iam_resources_across_accounts.all_roles_combined.s3.bucket"
            )
            s3_key = config.get(
                "cache_iam_resources_across_accounts.all_roles_combined.s3.file",
                "account_resource_cache/cache_all_roles_v1.json.gz",
            )
            try:
                all_roles = await retrieve_json_data_from_redis_or_s3(
                    redis_topic,
                    redis_data_type="hash",
                    s3_bucket=s3_bucket,
                    s3_key=s3_key,
                    json_object_hook=RoleAuthorizationsDecoder,
                    json_encoder=pydantic_encoder,
                    max_age=max_age,
                )
            except Exception as e:
                sentry_sdk.capture_exception()
                log.error(
                    {
                        "function": f"{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
                        "error": f"Error loading IAM roles. Returning empty list: {e}",
                    },
                    exc_info=True,
                )
                return []
            self._all_roles = list(all_roles.keys())
            self._all_roles_count = len(self._all_roles)
            self._all_roles_last_update = int(time.time())
        return self._all_roles

    async def all_roles(self, paginate=False, page=None, count=None):
        all_roles = await self.retrieve_all_roles()
        return all_roles

    async def number_roles(self) -> int:
        _ = await self.retrieve_all_roles()
        return self._all_roles_count

    async def determine_role_authorized_groups(self, account_id: str, role_name: str):
        arn = f"arn:aws:iam::{account_id}:role/{role_name.lower()}"
        reverse_mapping = await self.retrieve_reverse_authorization_mapping()
        groups = reverse_mapping.get(arn, [])
        return set(groups)

    async def determine_users_authorized_roles(self, user, groups, include_cli=False):
        authorization_mapping = await self.retrieve_credential_authorization_mapping()
        authorized_roles = set()
        user_mapping = authorization_mapping.get(user, [])
        if user_mapping:
            authorized_roles.update(user_mapping.authorized_roles)
            if include_cli:
                authorized_roles.update(user_mapping.authorized_roles_cli_only)
        for group in groups:
            group_mapping = authorization_mapping.get(group, [])
            if group_mapping:
                authorized_roles.update(group_mapping.authorized_roles)
                if include_cli:
                    authorized_roles.update(group_mapping.authorized_roles_cli_only)
        return sorted(authorized_roles)


async def generate_and_store_reverse_authorization_mapping(
    authorization_mapping: Dict[user_or_group, RoleAuthorizations]
) -> Dict[str, List[user_or_group]]:
    reverse_mapping = defaultdict(list)
    for identity, roles in authorization_mapping.items():
        for role in roles.authorized_roles:
            reverse_mapping[role.lower()].append(identity)
        for role in roles.authorized_roles_cli_only:
            reverse_mapping[role.lower()].append(identity)

    # Store in S3 and Redis
    redis_topic = config.get(
        "generate_and_store_reverse_authorization_mapping.redis_key",
        "REVERSE_AUTHORIZATION_MAPPING_V1",
    )
    s3_bucket = None
    s3_key = None
    if config.region == config.get("celery.active_region", config.region) or config.get(
        "environment"
    ) in ["dev", "test"]:
        s3_bucket = config.get(
            "generate_and_store_reverse_authorization_mapping.s3.bucket"
        )
        s3_key = config.get(
            "generate_and_store_reverse_authorization_mapping.s3.file",
            "reverse_authorization_mapping/reverse_authorization_mapping_v1.json.gz",
        )
    await store_json_results_in_redis_and_s3(
        reverse_mapping,
        redis_topic,
        s3_bucket=s3_bucket,
        s3_key=s3_key,
        json_encoder=pydantic_encoder,
    )
    return reverse_mapping


async def generate_and_store_credential_authorization_mapping() -> Dict[
    user_or_group, RoleAuthorizations
]:
    authorization_mapping: Dict[user_or_group, RoleAuthorizations] = {}

    if config.get("cloud_credential_authorization_mapping.role_tags.enabled", True):
        authorization_mapping = await RoleTagAuthorizationMappingGenerator().generate_credential_authorization_mapping(
            authorization_mapping
        )
    if config.get(
        "cloud_credential_authorization_mapping.dynamic_config.enabled", True
    ):
        authorization_mapping = await DynamicConfigAuthorizationMappingGenerator().generate_credential_authorization_mapping(
            authorization_mapping
        )
    if config.get(
        "cloud_credential_authorization_mapping.internal_plugin.enabled", False
    ):
        authorization_mapping = await InternalPluginAuthorizationMappingGenerator().generate_credential_authorization_mapping(
            authorization_mapping
        )

    # Store in S3 and Redis
    redis_topic = config.get(
        "generate_and_store_credential_authorization_mapping.redis_key",
        "CREDENTIAL_AUTHORIZATION_MAPPING_V1",
    )
    s3_bucket = None
    s3_key = None
    if config.region == config.get("celery.active_region", config.region) or config.get(
        "environment"
    ) in ["dev", "test"]:
        s3_bucket = config.get(
            "generate_and_store_credential_authorization_mapping.s3.bucket"
        )
        s3_key = config.get(
            "generate_and_store_credential_authorization_mapping.s3.file",
            "credential_authorization_mapping/credential_authorization_mapping_v1.json.gz",
        )
    await store_json_results_in_redis_and_s3(
        authorization_mapping,
        redis_topic,
        s3_bucket=s3_bucket,
        s3_key=s3_key,
        json_encoder=pydantic_encoder,
    )
    return authorization_mapping



import sentry_sdk

from consoleme.config import config
from consoleme.exceptions.exceptions import MustBeFte
from consoleme.handlers.base import BaseAPIV2Handler
from consoleme.lib.aws import get_scps_for_account_or_ou
from consoleme.models import Status2, WebResponse

log = config.get_logger()


class ServiceControlPolicyHandler(BaseAPIV2Handler):
    """
    Handler for /api/v2/service_control_policies/{accountNumberOrOuId}

    Returns Service Control Policies targeting specified account or OU
    """

    allowed_methods = ["GET"]

    async def get(self, identifier):
        if config.get("policy_editor.disallow_contractors", True) and self.contractor:
            if self.user not in config.get(
                "groups.can_bypass_contractor_restrictions", []
            ):
                raise MustBeFte("Only FTEs are authorized to view this page.")

        log_data = {
            "function": "ServiceControlPolicyHandler.get",
            "user": self.user,
            "message": "Retrieving service control policies for identifier",
            "identifier": identifier,
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
        }

        log.debug(log_data)
        try:
            scps = await get_scps_for_account_or_ou(identifier)
        except Exception as e:
            sentry_sdk.capture_exception()
            response = WebResponse(
                status=Status2.error, status_code=403, errors=[str(e)], data=[]
            )
            self.write(response.json())
            return
        response = WebResponse(
            status=Status2.success, status_code=200, data=scps.__root__
        )
        self.write(response.json())



import os
import sys
import threading
import time
from typing import Any, Optional

import boto3
import redis
import ujson as json
from asgiref.sync import sync_to_async
from redis.client import Redis

from consoleme.config import config
from consoleme.lib.plugins import get_plugin_by_name

if config.get("redis.use_redislite"):
    import tempfile

    import redislite

    if not config.get("redis.redis_lite.db_path"):
        default_redislite_db_path = tempfile.NamedTemporaryFile().name

region = config.region
log = config.get_logger()
stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()

automatically_backup_to_s3 = config.get(
    "redis.automatically_backup_to_s3.enabled", False
)
automatically_restore_from_s3 = config.get(
    "redis.automatically_restore_from_s3.enabled", False
)
s3 = boto3.resource("s3", **config.get("boto3.client_kwargs", {}))
s3_bucket = config.get("redis.automatically_backup_to_s3.bucket")
s3_folder = config.get("redis.automatically_backup_to_s3.folder")


# ToDo - Switch to Aioredis
class ConsoleMeRedis(redis.StrictRedis):
    """
    ConsoleMeRedis is a simple wrapper around redis.StrictRedis. It was created to allow Redis to be optional.
    If Redis settings are not defined in ConsoleMe's configuration, we "disable" redis. If Redis is disabled, calls to
    Redis will fail silently. If new Redis calls are added to ConsoleMe, they should be added to this class.

    ConsoleMeRedis also supports writing/retrieving data from S3 if the data is not retrievable from Redis
    """

    def __init__(self, *args, **kwargs):
        self.enabled = True
        if kwargs["host"] is None or kwargs["port"] is None or kwargs["db"] is None:
            self.enabled = False
        super(ConsoleMeRedis, self).__init__(*args, **kwargs)

    def get(self, *args, **kwargs):
        if not self.enabled:
            return None
        try:
            result = super(ConsoleMeRedis, self).get(*args, **kwargs)
        except redis.exceptions.ConnectionError as e:
            function = (
                f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
            )
            log.error(
                {
                    "function": function,
                    "message": "Unable to perform redis operation",
                    "key": args[0],
                    "error": e,
                },
                exc_info=True,
            )
            stats.count(f"{function}.error")
            result = None
        if not result and automatically_restore_from_s3:
            try:
                obj = s3.Object(s3_bucket, s3_folder + f"/{args[0]}")
                result = obj.get()["Body"].read().decode("utf-8")
            except s3.meta.client.exceptions.NoSuchKey:
                pass
            except Exception as e:
                function = f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
                log.error(
                    {
                        "function": function,
                        "message": "Unable to perform S3 operation",
                        "key": args[0],
                        "error": e,
                    },
                    exc_info=True,
                )
                stats.count(f"{function}.error")
        return result

    def set(self, *args, **kwargs):
        if not self.enabled:
            return False
        try:
            result = super(ConsoleMeRedis, self).set(*args, **kwargs)
        except redis.exceptions.ConnectionError as e:
            function = (
                f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
            )
            log.error(
                {
                    "function": function,
                    "message": "Unable to perform redis operation",
                    "key": args[0],
                    "error": e,
                },
                exc_info=True,
            )
            stats.count(f"{function}.error")
            result = None
        if automatically_backup_to_s3:
            try:
                obj = s3.Object(s3_bucket, s3_folder + f"/{args[0]}")
                obj.put(Body=str(args[1]))
            except Exception as e:
                function = f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
                log.error(
                    {
                        "function": function,
                        "message": "Unable to perform S3 operation",
                        "key": args[0],
                        "error": e,
                    },
                    exc_info=True,
                )
                stats.count(f"{function}.error")
        return result

    def setex(self, *args, **kwargs):
        if not self.enabled:
            return False
        # We do not currently support caching data in S3 with expiration (SETEX)
        try:
            result = super(ConsoleMeRedis, self).setex(*args, **kwargs)
        except redis.exceptions.ConnectionError as e:
            function = (
                f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
            )
            log.error(
                {
                    "function": function,
                    "message": "Unable to perform redis operation",
                    "key": args[0],
                    "error": e,
                },
                exc_info=True,
            )
            stats.count(f"{function}.error")
            result = None
        return result

    def hmset(self, *args, **kwargs):
        if not self.enabled:
            return False
        try:
            result = super(ConsoleMeRedis, self).hmset(*args, **kwargs)
        except redis.exceptions.ConnectionError as e:
            function = (
                f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
            )
            log.error(
                {
                    "function": function,
                    "message": "Unable to perform redis operation",
                    "key": args[0],
                    "error": e,
                },
                exc_info=True,
            )
            stats.count(f"{function}.error")
            result = None
        if automatically_backup_to_s3:
            try:
                obj = s3.Object(s3_bucket, s3_folder + f"/{args[0]}")
                # Write to S3 in a separate thread
                t = threading.Thread(
                    target=obj.put, kwargs={"Body": json.dumps(args[1])}
                )
                t.daemon = True
                t.start()
            except Exception as e:
                function = f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
                log.error(
                    {
                        "function": function,
                        "message": "Unable to perform S3 operation",
                        "key": args[0],
                        "error": e,
                    },
                    exc_info=True,
                )
                stats.count(f"{function}.error")
        return result

    def hset(self, *args, **kwargs):
        if not self.enabled:
            return False
        try:
            result = super(ConsoleMeRedis, self).hset(*args, **kwargs)
        except redis.exceptions.ConnectionError as e:
            function = (
                f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
            )
            log.error(
                {
                    "function": function,
                    "message": "Unable to perform redis operation",
                    "key": args[0],
                    "error": e,
                },
                exc_info=True,
            )
            stats.count(f"{function}.error")
            result = None
        if automatically_backup_to_s3:
            try:
                obj = s3.Object(s3_bucket, s3_folder + f"/{args[0]}")
                try:
                    current = json.loads(obj.get()["Body"].read().decode("utf-8"))
                    current[args[1]] = args[2]
                except:  # noqa
                    current = {args[1]: args[2]}
                t = threading.Thread(
                    target=obj.put, kwargs={"Body": json.dumps(current)}
                )
                t.daemon = True
                t.start()
            except Exception as e:
                function = f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
                log.error(
                    {
                        "function": function,
                        "message": "Unable to perform S3 operation",
                        "key": args[0],
                        "error": e,
                    },
                    exc_info=True,
                )
                stats.count(f"{function}.error")
        return result

    def hget(self, *args, **kwargs):
        if not self.enabled:
            return None
        try:
            result = super(ConsoleMeRedis, self).hget(*args, **kwargs)
        except redis.exceptions.ConnectionError as e:
            function = (
                f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
            )
            log.error(
                {
                    "function": function,
                    "message": "Unable to perform redis operation",
                    "key": args[0],
                    "error": e,
                },
                exc_info=True,
            )
            stats.count(f"{function}.error")
            result = None

        if not result and automatically_restore_from_s3:
            try:
                obj = s3.Object(s3_bucket, s3_folder + f"/{args[0]}")
                current = json.loads(obj.get()["Body"].read().decode("utf-8"))
                result = current.get(args[1])
                if result:
                    self.hset(args[0], args[1], result)
            except s3.meta.client.exceptions.NoSuchKey:
                pass
            except Exception as e:
                function = f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
                log.error(
                    {
                        "function": function,
                        "message": "Unable to perform S3 operation",
                        "key": args[0],
                        "error": e,
                    },
                    exc_info=True,
                )
                stats.count(f"{function}.error")
        return result

    def hmget(self, *args, **kwargs):
        if not self.enabled:
            return None
        try:
            result = super(ConsoleMeRedis, self).hmget(*args, **kwargs)
        except redis.exceptions.ConnectionError as e:
            function = (
                f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
            )
            log.error(
                {
                    "function": function,
                    "message": "Unable to perform redis operation",
                    "key": args[0],
                    "error": e,
                },
                exc_info=True,
            )
            stats.count(f"{function}.error")
            result = None
        return result

    def hgetall(self, *args, **kwargs):
        if not self.enabled:
            return None
        try:
            result = super(ConsoleMeRedis, self).hgetall(*args, **kwargs)
        except redis.exceptions.ConnectionError as e:
            function = (
                f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
            )
            log.error(
                {
                    "function": function,
                    "message": "Unable to perform redis operation",
                    "key": args[0],
                    "error": e,
                },
                exc_info=True,
            )
            stats.count(f"{function}.error")
            result = None
        if not result and automatically_restore_from_s3:
            try:
                obj = s3.Object(s3_bucket, s3_folder + f"/{args[0]}")
                result_j = obj.get()["Body"].read().decode("utf-8")
                result = json.loads(result_j)
                if result:
                    self.hmset(args[0], result)
            except s3.meta.client.exceptions.NoSuchKey:
                pass
            except Exception as e:
                function = f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
                log.error(
                    {
                        "function": function,
                        "message": "Unable to perform S3 operation",
                        "key": args[0],
                        "error": e,
                    },
                    exc_info=True,
                )
                stats.count(f"{function}.error")
        return result


class RedisHandler:
    def __init__(
        self,
        host: str = config.get(
            "redis.host.{}".format(region), config.get("redis.host.global", "localhost")
        ),
        port: int = config.get("redis.port", 6379),
        db: int = config.get("redis.db", 0),
    ) -> None:
        self.red = None
        self.host = host
        self.port = port
        self.db = db
        self.enabled = True
        if self.host is None or self.port is None or self.db is None:
            self.enabled = False

    async def redis(self, db: int = 0) -> Redis:
        if config.get("redis.use_redislite"):
            REDIS_DB_PATH = os.path.join(
                config.get("redis.redislite.db_path", default_redislite_db_path)
            )
            return redislite.StrictRedis(REDIS_DB_PATH, decode_responses=True)
        self.red = await sync_to_async(ConsoleMeRedis)(
            host=self.host,
            port=self.port,
            db=self.db,
            charset="utf-8",
            decode_responses=True,
        )
        return self.red

    def redis_sync(self, db: int = 0) -> Redis:
        if config.get("redis.use_redislite"):
            REDIS_DB_PATH = os.path.join(
                config.get("redis.redislite.db_path", default_redislite_db_path)
            )
            return redislite.StrictRedis(REDIS_DB_PATH, decode_responses=True)
        self.red = ConsoleMeRedis(
            host=self.host,
            port=self.port,
            db=self.db,
            charset="utf-8",
            decode_responses=True,
        )
        return self.red


async def redis_get(key: str, default: Optional[str] = None) -> Optional[str]:
    red = await RedisHandler().redis()
    v = await sync_to_async(red.get)(key)
    if not v:
        return default
    return v


async def redis_hgetall(key: str, default=None):
    red = await RedisHandler().redis()
    v = await sync_to_async(red.hgetall)(key)
    if not v:
        return default
    return v


async def redis_hget(name: str, key: str, default=None):
    red = await RedisHandler().redis()
    v = await sync_to_async(red.hget)(name, key)
    if not v:
        return default
    return v


def redis_get_sync(key: str, default: None = None) -> Optional[str]:
    red = RedisHandler().redis_sync()
    try:
        v = red.get(key)
    except redis.exceptions.ConnectionError:
        v = None
    if not v:
        return default
    return v


async def redis_hsetex(name: str, key: str, value: Any, expiration_seconds: int):
    """
    Lazy way to set Redis hash keys with an expiration. Warning: Entries set here only get deleted when redis_hgetex
    is called on an expired key.

    :param name: Redis key
    :param key: Hash key
    :param value: Hash value
    :param expiration_seconds: Number of seconds to consider entry expired
    :return:
    """
    expiration = int(time.time()) + expiration_seconds
    red = await RedisHandler().redis()
    v = await sync_to_async(red.hset)(
        name, key, json.dumps({"value": value, "ttl": expiration})
    )
    return v


async def redis_hgetex(name: str, key: str, default=None):
    """
    Lazy way to retrieve an entry from a Redis Hash, and delete it if it's due to expire.

    :param name:
    :param key:
    :param default:
    :return:
    """
    red = await RedisHandler().redis()
    if not red.exists(name):
        return default
    result_j = await sync_to_async(red.hget)(name, key)
    if not result_j:
        return default
    result = json.loads(result_j)
    if int(time.time()) > result["ttl"]:
        red.hdel(name, key)
        return default
    return result["value"]



import asyncio
import sys
import time
import uuid

import sentry_sdk
import ujson as json
from policy_sentry.util.arns import parse_arn
from pydantic import ValidationError

from consoleme.celery_tasks.celery_tasks import app as celery_app
from consoleme.config import config
from consoleme.exceptions.exceptions import (
    InvalidRequestParameter,
    MustBeFte,
    NoMatchingRequest,
    ResourceNotFound,
    Unauthorized,
)
from consoleme.handlers.base import BaseAPIV2Handler, BaseHandler
from consoleme.lib.auth import can_admin_policies
from consoleme.lib.aws import get_resource_account
from consoleme.lib.cache import retrieve_json_data_from_redis_or_s3
from consoleme.lib.dynamo import UserDynamoHandler
from consoleme.lib.generic import filter_table, write_json_error
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.lib.policies import (
    can_move_back_to_pending_v2,
    can_update_cancel_requests_v2,
    get_url_for_resource,
    should_auto_approve_policy_v2,
)
from consoleme.lib.slack import send_slack_notification_new_request
from consoleme.lib.timeout import Timeout
from consoleme.lib.v2.requests import (
    generate_request_from_change_model_array,
    get_request_url,
    is_request_eligible_for_auto_approval,
    parse_and_apply_policy_request_modification,
    populate_cross_account_resource_policies,
    populate_old_managed_policies,
    populate_old_policies,
)
from consoleme.models import (
    CommentModel,
    DataTableResponse,
    ExtendedRequestModel,
    PolicyRequestModificationRequestModel,
    RequestCreationModel,
    RequestCreationResponse,
    RequestStatus,
)

stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
log = config.get_logger()
aws = get_plugin_by_name(config.get("plugins.aws", "default_aws"))()


class RequestHandler(BaseAPIV2Handler):
    """Handler for /api/v2/request

    Allows for creation of a request.
    """

    allowed_methods = ["POST"]

    def on_finish(self) -> None:
        if self.request.method != "POST":
            return
        celery_app.send_task(
            "consoleme.celery_tasks.celery_tasks.cache_policy_requests"
        )
        celery_app.send_task(
            "consoleme.celery_tasks.celery_tasks.cache_credential_authorization_mapping"
        )

    async def post(self):
        """
        POST /api/v2/request

        Request example JSON: (Request Schema is RequestCreationModel in models.py)

        {
          "changes": {
            "changes": [
              {
                "principal": {
                    "principal_arn": "arn:aws:iam::123456789012:role/curtisTestRole1",
                    "principal_type": "AwsResource"
                },
                "change_type": "inline_policy",
                "action": "attach",
                "policy": {
                  "policy_document": {
                    "Version": "2012-10-17",
                    "Statement": [
                      {
                        "Action": [
                          "s3:ListMultipartUploadParts*",
                          "s3:ListBucket"
                        ],
                        "Effect": "Allow",
                        "Resource": [
                          "arn:aws:s3:::curtis-nflx-test/*",
                          "arn:aws:s3:::curtis-nflx-test"
                        ],
                        "Sid": "cmccastrapel159494014dsd1shak"
                      },
                      {
                        "Action": [
                          "ec2:describevolumes",
                          "ec2:detachvolume",
                          "ec2:describelicenses",
                          "ec2:AssignIpv6Addresses",
                          "ec2:reportinstancestatus"
                        ],
                        "Effect": "Allow",
                        "Resource": [
                          "*"
                        ],
                        "Sid": "cmccastrapel1594940141hlvvv"
                      },
                      {
                        "Action": [
                          "sts:AssumeRole"
                        ],
                        "Effect": "Allow",
                        "Resource": [
                          "arn:aws:iam::123456789012:role/curtisTestInstanceProfile"
                        ],
                        "Sid": "cmccastrapel1596483596easdits"
                      }
                    ]
                  }
                }
              },
              {
                "principal": {
                    "principal_arn": "arn:aws:iam::123456789012:role/curtisTestRole1",
                    "principal_type": "AwsResource"
                },
                "change_type": "assume_role_policy",
                "policy": {
                  "policy_document": {
                    "Statement": [
                      {
                        "Action": "sts:AssumeRole",
                        "Effect": "Allow",
                        "Principal": {
                          "AWS": "arn:aws:iam::123456789012:role/consolemeInstanceProfile"
                        },
                        "Sid": "AllowConsoleMeProdAssumeRolses"
                      }
                    ],
                    "Version": "2012-10-17"
                  }
                }
              },
              {
                "principal": {
                    "principal_arn": "arn:aws:iam::123456789012:role/curtisTestRole1",
                    "principal_type": "AwsResource"
                },
                "change_type": "managed_policy",
                "policy_name": "ApiProtect",
                "action": "attach",
                "arn": "arn:aws:iam::123456789012:policy/ApiProtect"
              },
              {
                "principal": {
                    "principal_arn": "arn:aws:iam::123456789012:role/curtisTestRole1",
                    "principal_type": "AwsResource"
                },
                "change_type": "managed_policy",
                "policy_name": "TagProtect",
                "action": "detach",
                "arn": "arn:aws:iam::123456789012:policy/TagProtect"
              },
              {
                "principal": {
                    "principal_arn": "arn:aws:iam::123456789012:role/curtisTestRole1",
                    "principal_type": "AwsResource"
                },
                "change_type": "inline_policy",
                "policy_name": "random_policy254",
                "action": "attach",
                "policy": {
                  "policy_document": {
                    "Version": "2012-10-17",
                    "Statement": [
                      {
                        "Action": [
                          "ec2:AssignIpv6Addresses"
                        ],
                        "Effect": "Allow",
                        "Resource": [
                          "*"
                        ],
                        "Sid": "cmccastrapel1594940141shakabcd"
                      }
                    ]
                  }
                }
              }
            ]
          },
          "justification": "testing this out.",
          "admin_auto_approve": false
        }

        Response example JSON: (Response Schema is RequestCreationResponse in models.py)

        {
            "errors": 1,
            "request_created": true,
            "request_id": "0c9fb298-c8ea-4d50-917c-3212da07b3ad",
            "action_results": [
                {
                    "status": "success",
                    "message": "Success description"
                },
                {
                    "status": "error",
                    "message": "Error description"
                }
            ]
        }


        """

        if config.get("policy_editor.disallow_contractors", True) and self.contractor:
            if self.user not in config.get(
                "groups.can_bypass_contractor_restrictions", []
            ):
                raise MustBeFte("Only FTEs are authorized to view this page.")

        tags = {"user": self.user}
        stats.count("RequestHandler.post", tags=tags)
        log_data = {
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "user": self.user,
            "message": "Create request initialization",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
            "ip": self.ip,
            "admin_auto_approved": False,
            "probe_auto_approved": False,
        }
        log.debug(log_data)
        try:
            # Validate the model
            changes = RequestCreationModel.parse_raw(self.request.body)
            extended_request = await generate_request_from_change_model_array(
                changes, self.user
            )
            log_data["request"] = extended_request.dict()
            log.debug(log_data)

            if changes.dry_run:
                response = RequestCreationResponse(
                    errors=0, request_created=False, extended_request=extended_request
                )
                self.write(response.json())
                await self.finish()
                return

            admin_approved = False
            approval_probe_approved = False

            if extended_request.principal.principal_type == "AwsResource":
                # TODO: Provide a note to the requester that admin_auto_approve will apply the requested policies only.
                # It will not automatically apply generated policies. The administrative user will need to visit
                # the policy request page to do this manually.
                if changes.admin_auto_approve:
                    # make sure user is allowed to use admin_auto_approve
                    can_manage_policy_request = (
                        can_admin_policies(self.user, self.groups),
                    )
                    if can_manage_policy_request:
                        extended_request.request_status = RequestStatus.approved
                        admin_approved = True
                        extended_request.reviewer = self.user
                        self_approval_comment = CommentModel(
                            id=str(uuid.uuid4()),
                            timestamp=int(time.time()),
                            user_email=self.user,
                            user=extended_request.requester_info,
                            last_modified=int(time.time()),
                            text=f"Self-approved by admin: {self.user}",
                        )
                        extended_request.comments.append(self_approval_comment)
                        log_data["admin_auto_approved"] = True
                        log_data["request"] = extended_request.dict()
                        log.debug(log_data)
                        stats.count(
                            f"{log_data['function']}.post.admin_auto_approved",
                            tags={"user": self.user},
                        )
                    else:
                        # someone is trying to use admin bypass without being an admin, don't allow request to proceed
                        stats.count(
                            f"{log_data['function']}.post.unauthorized_admin_bypass",
                            tags={"user": self.user},
                        )
                        log_data[
                            "message"
                        ] = "Unauthorized user trying to use admin bypass"
                        log.error(log_data)
                        await write_json_error("Unauthorized", obj=self)
                        return
                else:
                    # If admin auto approve is false, check for auto-approve probe eligibility
                    is_eligible_for_auto_approve_probe = (
                        await is_request_eligible_for_auto_approval(
                            extended_request, self.user
                        )
                    )
                    # If we have only made requests that are eligible for auto-approval probe, check against them
                    if is_eligible_for_auto_approve_probe:
                        should_auto_approve_request = (
                            await should_auto_approve_policy_v2(
                                extended_request, self.user, self.groups
                            )
                        )
                        if should_auto_approve_request["approved"]:
                            extended_request.request_status = RequestStatus.approved
                            approval_probe_approved = True
                            stats.count(
                                f"{log_data['function']}.probe_auto_approved",
                                tags={"user": self.user},
                            )
                            approving_probes = []
                            for approving_probe in should_auto_approve_request[
                                "approving_probes"
                            ]:
                                approving_probe_comment = CommentModel(
                                    id=str(uuid.uuid4()),
                                    timestamp=int(time.time()),
                                    user_email=f"Auto-Approve Probe: {approving_probe['name']}",
                                    last_modified=int(time.time()),
                                    text=(
                                        f"Policy {approving_probe['policy']} auto-approved by probe: "
                                        f"{approving_probe['name']}"
                                    ),
                                )
                                extended_request.comments.append(
                                    approving_probe_comment
                                )
                                approving_probes.append(approving_probe["name"])
                            extended_request.reviewer = (
                                f"Auto-Approve Probe: {','.join(approving_probes)}"
                            )
                            log_data["probe_auto_approved"] = True
                            log_data["request"] = extended_request.dict()
                            log.debug(log_data)

            dynamo = UserDynamoHandler(self.user)
            request = await dynamo.write_policy_request_v2(extended_request)
            log_data["message"] = "New request created in Dynamo"
            log_data["request"] = extended_request.dict()
            log_data["dynamo_request"] = request
            log.debug(log_data)
        except (InvalidRequestParameter, ValidationError) as e:
            log_data["message"] = "Validation Exception"
            log.error(log_data, exc_info=True)
            stats.count(
                f"{log_data['function']}.validation_exception", tags={"user": self.user}
            )
            self.write_error(400, message="Error validating input: " + str(e))
            if config.get("development"):
                raise
            return
        except Exception as e:
            log_data["message"] = "Unknown Exception occurred while parsing request"
            log.error(log_data, exc_info=True)
            stats.count(f"{log_data['function']}.exception", tags={"user": self.user})
            sentry_sdk.capture_exception(tags={"user": self.user})
            self.write_error(500, message="Error parsing request: " + str(e))
            if config.get("development"):
                raise
            return

        request_url = await get_request_url(extended_request)

        # If here, request has been successfully created
        response = RequestCreationResponse(
            errors=0,
            request_created=True,
            request_id=extended_request.id,
            request_url=request_url,
            action_results=[],
            extended_request=extended_request,
        )

        # If approved is true due to an auto-approval probe or admin auto-approval, apply the non-autogenerated changes
        if extended_request.request_status == RequestStatus.approved:
            for change in extended_request.changes.changes:
                if change.autogenerated:
                    continue
                policy_request_modification_model = (
                    PolicyRequestModificationRequestModel.parse_obj(
                        {
                            "modification_model": {
                                "command": "apply_change",
                                "change_id": change.id,
                            }
                        }
                    )
                )
                policy_apply_response = (
                    await parse_and_apply_policy_request_modification(
                        extended_request,
                        policy_request_modification_model,
                        self.user,
                        self.groups,
                        int(time.time()),
                        approval_probe_approved,
                    )
                )
                response.errors = policy_apply_response.errors
                response.action_results = policy_apply_response.action_results

            # Update in dynamo
            await dynamo.write_policy_request_v2(extended_request)
            account_id = await get_resource_account(
                extended_request.principal.principal_arn
            )

            # Force a refresh of the role in Redis/DDB
            arn_parsed = parse_arn(extended_request.principal.principal_arn)
            if arn_parsed["service"] == "iam" and arn_parsed["resource"] == "role":
                await aws.fetch_iam_role(
                    account_id,
                    extended_request.principal.principal_arn,
                    force_refresh=True,
                )
            log_data["request"] = extended_request.dict()
            log_data["message"] = "Applied changes based on approved request"
            log_data["response"] = response.dict()
            log.debug(log_data)

        await aws.send_communications_new_policy_request(
            extended_request, admin_approved, approval_probe_approved
        )
        await send_slack_notification_new_request(
            extended_request, admin_approved, approval_probe_approved
        )
        self.write(response.json())
        await self.finish()
        return


class RequestsHandler(BaseAPIV2Handler):
    """Handler for /api/v2/requests

    Api endpoint to list and filter policy requests.
    """

    allowed_methods = ["POST"]

    async def post(self):
        """
        POST /api/v2/requests
        """
        arguments = {k: self.get_argument(k) for k in self.request.arguments}
        markdown = arguments.get("markdown")
        cache_key = config.get(
            "cache_all_policy_requests.redis_key", "ALL_POLICY_REQUESTS"
        )
        s3_bucket = config.get("cache_policy_requests.s3.bucket")
        s3_key = config.get(
            "cache_policy_requests.s3.file",
            "policy_requests/all_policy_requests_v1.json.gz",
        )
        arguments = json.loads(self.request.body)
        filters = arguments.get("filters")
        # TODO: Add server-side sorting
        # sort = arguments.get("sort")
        limit = arguments.get("limit", 1000)
        tags = {"user": self.user}
        stats.count("RequestsHandler.post", tags=tags)
        log_data = {
            "function": "RequestsHandler.post",
            "user": self.user,
            "message": "Writing requests",
            "limit": limit,
            "filters": filters,
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
        }
        log.debug(log_data)
        requests = await retrieve_json_data_from_redis_or_s3(
            cache_key, s3_bucket=s3_bucket, s3_key=s3_key
        )

        total_count = len(requests)

        if filters:
            try:
                with Timeout(seconds=5):
                    for filter_key, filter_value in filters.items():
                        requests = await filter_table(
                            filter_key, filter_value, requests
                        )
            except TimeoutError:
                self.write("Query took too long to run. Check your filter.")
                await self.finish()
                raise

        if markdown:
            requests_to_write = []
            for request in requests[0:limit]:
                principal_arn = request.get("principal", {}).get("principal_arn", "")
                url = request.get("principal", {}).get("resource_url", "")
                resource_name = principal_arn
                if "/" in resource_name:
                    resource_name = resource_name.split("/")[-1]
                if not resource_name:
                    resource_name = request.get("principal", {}).get(
                        "resource_identifier"
                    )

                if principal_arn and principal_arn.count(":") == 5 and not url:

                    region = principal_arn.split(":")[3]
                    service_type = principal_arn.split(":")[2]
                    account_id = principal_arn.split(":")[4]
                    if request.get("principal", {}).get("principal_arn"):
                        try:
                            url = await get_url_for_resource(
                                principal_arn,
                                service_type,
                                account_id,
                                region,
                                resource_name,
                            )
                        except ResourceNotFound:
                            pass
                # Convert request_id and role ARN to link
                request_url = request.get("extended_request", {}).get("request_url")
                if not request_url:
                    request_url = f"/policies/request/{request['request_id']}"
                request["request_id"] = f"[{request['request_id']}]({request_url})"
                if url:
                    request["arn"] = f"[{principal_arn or resource_name}]({url})"
                requests_to_write.append(request)
        else:
            requests_to_write = requests[0:limit]
        filtered_count = len(requests_to_write)
        res = DataTableResponse(
            totalCount=total_count, filteredCount=filtered_count, data=requests_to_write
        )
        self.write(res.json())
        return


class RequestDetailHandler(BaseAPIV2Handler):
    """Handler for /api/v2/requests/{request_id}

    Allows read and update access to a specific request.
    """

    allowed_methods = ["GET", "PUT"]

    def on_finish(self) -> None:
        if self.request.method != "PUT":
            return
        celery_app.send_task(
            "consoleme.celery_tasks.celery_tasks.cache_policy_requests"
        )
        celery_app.send_task(
            "consoleme.celery_tasks.celery_tasks.cache_credential_authorization_mapping"
        )

    async def _get_extended_request(self, request_id, log_data):
        dynamo = UserDynamoHandler(self.user)
        requests = await dynamo.get_policy_requests(request_id=request_id)
        if len(requests) == 0:
            log_data["message"] = "Request with that ID not found"
            log.warning(log_data)
            stats.count(f"{log_data['function']}.not_found", tags={"user": self.user})
            raise NoMatchingRequest(log_data["message"])
        if len(requests) > 1:
            log_data["message"] = "Multiple requests with that ID found"
            log.error(log_data)
            stats.count(
                f"{log_data['function']}.multiple_requests_found",
                tags={"user": self.user},
            )
            raise InvalidRequestParameter(log_data["message"])
        request = requests[0]

        if request.get("version") != "2":
            # Request format is not compatible with this endpoint version
            raise InvalidRequestParameter("Request with that ID is not a v2 request")

        extended_request = ExtendedRequestModel.parse_obj(
            request.get("extended_request")
        )
        return extended_request, request.get("last_updated")

    async def get(self, request_id):
        """
        GET /api/v2/requests/{request_id}
        """
        tags = {"user": self.user}
        stats.count("RequestDetailHandler.get", tags=tags)
        log_data = {
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "user": self.user,
            "message": "Get request details",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
            "policy_request_id": request_id,
        }
        log.debug(log_data)

        if config.get("policy_editor.disallow_contractors", True) and self.contractor:
            if self.user not in config.get(
                "groups.can_bypass_contractor_restrictions", []
            ):
                self.write_error(
                    403, message="Only FTEs are authorized to view this page."
                )
                return

        try:
            extended_request, last_updated = await self._get_extended_request(
                request_id, log_data
            )
        except InvalidRequestParameter as e:
            sentry_sdk.capture_exception(tags={"user": self.user})
            self.write_error(400, message="Error validating input: " + str(e))
            return
        except NoMatchingRequest as e:
            sentry_sdk.capture_exception(tags={"user": self.user})
            self.write_error(404, message="Error getting request:" + str(e))
            return
        # Run these tasks concurrently.
        concurrent_results = await asyncio.gather(
            populate_old_policies(extended_request, self.user),
            populate_cross_account_resource_policies(extended_request, self.user),
            populate_old_managed_policies(extended_request, self.user),
        )
        extended_request = concurrent_results[0]

        populate_cross_account_resource_policies_result = concurrent_results[1]

        if populate_cross_account_resource_policies_result["changed"]:
            extended_request = populate_cross_account_resource_policies_result[
                "extended_request"
            ]
            # Update in dynamo with the latest resource policy changes
            dynamo = UserDynamoHandler(self.user)
            updated_request = await dynamo.write_policy_request_v2(extended_request)
            last_updated = updated_request.get("last_updated")

        populate_old_managed_policies_result = concurrent_results[2]

        if populate_old_managed_policies_result["changed"]:
            extended_request = populate_old_managed_policies_result["extended_request"]
            # Update in dynamo with the latest resource policy changes
            dynamo = UserDynamoHandler(self.user)
            updated_request = await dynamo.write_policy_request_v2(extended_request)
            last_updated = updated_request.get("last_updated")

        can_approve_reject = can_admin_policies(self.user, self.groups)
        can_update_cancel = await can_update_cancel_requests_v2(
            extended_request.requester_email, self.user, self.groups
        )
        can_move_back_to_pending = await can_move_back_to_pending_v2(
            extended_request, last_updated, self.user, self.groups
        )

        # In the future request_specific_config will have specific approvers for specific changes based on ABAC
        request_specific_config = {
            "can_approve_reject": can_approve_reject,
            "can_update_cancel": can_update_cancel,
            "can_move_back_to_pending": can_move_back_to_pending,
        }

        template = None
        # Force a refresh of the role in Redis/DDB
        arn_parsed = parse_arn(extended_request.principal.principal_arn)
        if arn_parsed["service"] == "iam" and arn_parsed["resource"] == "role":
            iam_role = await aws.fetch_iam_role(
                arn_parsed["account"], extended_request.principal.principal_arn
            )
            template = iam_role.get("templated")
        response = {
            "request": extended_request.json(),
            "last_updated": last_updated,
            "request_config": request_specific_config,
            "template": template,
        }

        self.write(response)

    async def put(self, request_id):
        """
        PUT /api/v2/requests/{request_id}
        """
        tags = {"user": self.user}
        stats.count("RequestDetailHandler.put", tags=tags)
        log_data = {
            "function": "RequestDetailHandler.put",
            "user": self.user,
            "message": "Incoming request",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
            "policy_request_id": request_id,
        }
        log.debug(log_data)

        if config.get("policy_editor.disallow_contractors", True) and self.contractor:
            if self.user not in config.get(
                "groups.can_bypass_contractor_restrictions", []
            ):
                raise MustBeFte("Only FTEs are authorized to view this page.")

        try:
            # Validate the request body
            request_changes = PolicyRequestModificationRequestModel.parse_raw(
                self.request.body
            )
            log_data["message"] = "Parsed request body"
            log_data["request"] = request_changes.dict()
            log.debug(log_data)

            extended_request, last_updated = await self._get_extended_request(
                request_id, log_data
            )
            response = await parse_and_apply_policy_request_modification(
                extended_request, request_changes, self.user, self.groups, last_updated
            )

        except (NoMatchingRequest, InvalidRequestParameter, ValidationError) as e:
            log_data["message"] = "Validation Exception"
            log.error(log_data, exc_info=True)
            sentry_sdk.capture_exception(tags={"user": self.user})
            stats.count(
                f"{log_data['function']}.validation_exception", tags={"user": self.user}
            )
            self.write_error(400, message="Error validating input: " + str(e))
            if config.get("development"):
                raise
            return
        except Unauthorized as e:
            log_data["message"] = "Unauthorized"
            log.error(log_data, exc_info=True)
            sentry_sdk.capture_exception(tags={"user": self.user})
            stats.count(
                f"{log_data['function']}.unauthorized", tags={"user": self.user}
            )
            self.write_error(403, message=str(e))
            if config.get("development"):
                raise
            return
        self.write(response.json())
        await self.finish()
        return


class RequestsPageConfigHandler(BaseHandler):
    async def get(self):
        """
        /requests_page_config
        ---
        get:
            description: Retrieve Requests Page Configuration
            responses:
                200:
                    description: Returns Requests Page Configuration
        """
        default_configuration = {
            "pageName": "Requests",
            "pageDescription": "View all IAM policy requests created through ConsoleMe",
            "tableConfig": {
                "expandableRows": True,
                "dataEndpoint": "/api/v2/requests?markdown=true",
                "sortable": False,
                "totalRows": 200,
                "rowsPerPage": 50,
                "serverSideFiltering": True,
                "allowCsvExport": True,
                "allowJsonExport": True,
                "columns": [
                    {
                        "placeholder": "Username",
                        "key": "username",
                        "type": "input",
                        "style": {"width": "100px"},
                    },
                    {
                        "placeholder": "Arn",
                        "key": "arn",
                        "type": "link",
                        "style": {"whiteSpace": "normal", "wordBreak": "break-all"},
                        "width": 3,
                    },
                    {
                        "placeholder": "Request Time",
                        "key": "request_time",
                        "type": "daterange",
                    },
                    {
                        "placeholder": "Status",
                        "key": "status",
                        "type": "dropdown",
                        "style": {"width": "90px"},
                    },
                    {
                        "placeholder": "Request ID",
                        "key": "request_id",
                        "type": "link",
                        "style": {"whiteSpace": "normal", "wordBreak": "break-all"},
                        "width": 2,
                    },
                ],
            },
        }

        table_configuration = config.get(
            "RequestsTableConfigHandler.configuration", default_configuration
        )

        self.write(table_configuration)



import asyncio
import os
import sys
import time
import uuid
import zlib
from collections import defaultdict
from datetime import datetime

# used as a placeholder for empty SID to work around this:
# https://github.com/aws/aws-sdk-js/issues/833
from decimal import Decimal
from typing import Any, Dict, List, Optional, Union

import bcrypt
import boto3
import sentry_sdk
import simplejson as json
import yaml
from asgiref.sync import sync_to_async
from boto3.dynamodb.conditions import Key
from boto3.dynamodb.types import Binary  # noqa
from cloudaux import get_iso_string
from cloudaux.aws.sts import boto3_cached_conn
from retrying import retry
from tenacity import Retrying, stop_after_attempt, wait_fixed

from consoleme.config import config
from consoleme.exceptions.exceptions import (
    DataNotRetrievable,
    NoExistingRequest,
    NoMatchingRequest,
    PendingRequestAlreadyExists,
)
from consoleme.lib.crypto import Crypto
from consoleme.lib.password import wait_after_authentication_failure
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.lib.redis import RedisHandler
from consoleme.models import AuthenticationResponse, ExtendedRequestModel

DYNAMO_EMPTY_STRING = "---DYNAMO-EMPTY-STRING---"

# We need to import Decimal to eval the request. This Decimal usage is to prevent lint errors on importing the unused
# Decimal module.
DYNAMODB_EMPTY_DECIMAL = Decimal(0)

POSSIBLE_STATUSES = config.get(
    "possible_statuses",
    ["pending", "approved", "rejected", "cancelled", "expired", "removed"],
)

stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
log = config.get_logger("consoleme")
crypto = Crypto()
red = RedisHandler().redis_sync()


class BaseDynamoHandler:
    """Base class for interacting with DynamoDB."""

    def _get_dynamo_table(self, table_name):
        function: str = (
            f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
        )
        try:
            # call sts_conn with my client and pass in forced_client
            if config.get("dynamodb_server"):
                resource = boto3.resource(
                    "dynamodb",
                    region_name=config.region,
                    endpoint_url=config.get(
                        "dynamodb_server",
                        config.get("boto3.client_kwargs.endpoint_url"),
                    ),
                )
            else:
                resource = boto3_cached_conn(
                    "dynamodb",
                    service_type="resource",
                    account_number=config.get("aws.account_number"),
                    session_name=config.get("application_name", "consoleme"),
                    region=config.region,
                    client_kwargs=config.get("boto3.client_kwargs", {}),
                )
            table = resource.Table(table_name)
        except Exception as e:
            log.error({"function": function, "error": e}, exc_info=True)
            stats.count(f"{function}.exception")
            return None
        else:
            return table

    def _data_from_dynamo_replace(
        self,
        obj: Union[
            List[Dict[str, Union[Decimal, str]]],
            Dict[str, Union[Decimal, str]],
            str,
            Decimal,
        ],
    ) -> Union[int, Dict[str, Union[int, str]], str, List[Dict[str, Union[int, str]]]]:
        """Traverse a potentially nested object and replace all Dynamo placeholders with actual empty strings

        Args:
            obj (object)

        Returns:
            object: Object with original empty strings

        """
        if isinstance(obj, dict):
            for k in ["aws:rep:deleting", "aws:rep:updateregion", "aws:rep:updatetime"]:
                if k in obj.keys():
                    del obj[k]
            return {k: self._data_from_dynamo_replace(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [self._data_from_dynamo_replace(elem) for elem in obj]
        else:
            if isinstance(obj, Binary):
                obj = obj.value
            if str(obj) == DYNAMO_EMPTY_STRING:
                obj = ""
            elif isinstance(obj, Decimal):
                obj = int(obj)
            return obj

    def _data_to_dynamo_replace(self, obj: Any) -> Any:
        """Traverse a potentially nested object and replace all instances of an empty string with a placeholder

        Args:
            obj (object)

        Returns:
            object: Object with Dynamo friendly empty strings

        """
        if isinstance(obj, dict):
            for k in ["aws:rep:deleting", "aws:rep:updateregion", "aws:rep:updatetime"]:
                if k in obj.keys():
                    del obj[k]
            return {k: self._data_to_dynamo_replace(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [self._data_to_dynamo_replace(elem) for elem in obj]
        else:
            if isinstance(obj, Binary):
                return obj
            if str(obj) == "":
                obj = DYNAMO_EMPTY_STRING
            elif type(obj) in [float, int]:
                obj = Decimal(str(obj))
            elif isinstance(obj, datetime):
                obj = Decimal(str(obj.timestamp()))
            return obj

    def parallel_write_table(self, table, data, overwrite_by_pkeys=None):
        if not overwrite_by_pkeys:
            overwrite_by_pkeys = []
        with table.batch_writer(overwrite_by_pkeys=overwrite_by_pkeys) as batch:
            for item in data:
                for attempt in Retrying(stop=stop_after_attempt(3), wait=wait_fixed(2)):
                    with attempt:
                        batch.put_item(Item=self._data_to_dynamo_replace(item))

    def parallel_scan_table(
        self,
        table,
        total_threads=os.cpu_count(),
        dynamodb_kwargs: Optional[Dict[str, Any]] = None,
    ):
        if not dynamodb_kwargs:
            dynamodb_kwargs = {}

        async def _scan_segment(segment, total_segments):
            response = table.scan(
                Segment=segment, TotalSegments=total_segments, **dynamodb_kwargs
            )
            items = response.get("Items", [])

            while "LastEvaluatedKey" in response:
                response = table.scan(
                    ExclusiveStartKey=response["LastEvaluatedKey"],
                    Segment=segment,
                    TotalSegments=total_segments,
                    **dynamodb_kwargs,
                )
                items.extend(self._data_from_dynamo_replace(response["Items"]))

            return items

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        tasks = []
        for i in range(total_threads):
            task = asyncio.ensure_future(_scan_segment(i, total_threads))
            tasks.append(task)

        results = loop.run_until_complete(asyncio.gather(*tasks))
        items = []
        for result in results:
            items.extend(result)
        return items

    async def parallel_scan_table_async(
        self,
        table,
        total_threads=os.cpu_count(),
        dynamodb_kwargs: Optional[Dict[str, Any]] = None,
    ):
        if not dynamodb_kwargs:
            dynamodb_kwargs = {}

        async def _scan_segment(segment, total_segments):
            response = table.scan(
                Segment=segment, TotalSegments=total_segments, **dynamodb_kwargs
            )
            items = response.get("Items", [])

            while "LastEvaluatedKey" in response:
                response = table.scan(
                    ExclusiveStartKey=response["LastEvaluatedKey"],
                    Segment=segment,
                    TotalSegments=total_segments,
                    **dynamodb_kwargs,
                )
                items.extend(self._data_from_dynamo_replace(response["Items"]))

            return items

        tasks = []
        for i in range(total_threads):
            task = asyncio.ensure_future(_scan_segment(i, total_threads))
            tasks.append(task)

        results = await asyncio.gather(*tasks)
        items = []
        for result in results:
            items.extend(result)
        return items


class UserDynamoHandler(BaseDynamoHandler):
    def __init__(self, user_email: Optional[str] = None) -> None:
        try:
            self.requests_table = self._get_dynamo_table(
                config.get("aws.requests_dynamo_table", "consoleme_requests_global")
            )
            self.users_table = self._get_dynamo_table(
                config.get("aws.users_dynamo_table", "consoleme_users_global")
            )
            self.group_log = self._get_dynamo_table(
                config.get("aws.group_log_dynamo_table", "consoleme_audit_global")
            )
            self.dynamic_config = self._get_dynamo_table(
                config.get("aws.group_log_dynamo_table", "consoleme_config_global")
            )
            self.policy_requests_table = self._get_dynamo_table(
                config.get(
                    "aws.policy_requests_dynamo_table", "consoleme_policy_requests"
                )
            )
            self.api_health_roles_table = self._get_dynamo_table(
                config.get(
                    "aws.api_health_apps_table_dynamo_table",
                    "consoleme_api_health_apps",
                )
            )
            self.resource_cache_table = self._get_dynamo_table(
                config.get(
                    "aws.resource_cache_dynamo_table", "consoleme_resource_cache"
                )
            )
            self.cloudtrail_table = self._get_dynamo_table(
                config.get("aws.cloudtrail_table", "consoleme_cloudtrail")
            )

            self.notifications_table = self._get_dynamo_table(
                config.get("aws.notifications_table", "consoleme_notifications")
            )

            if user_email:
                self.user = self.get_or_create_user(user_email)
                self.affected_user = self.user
        except Exception:
            if config.get("development"):
                log.error(
                    "Unable to connect to Dynamo. Trying to set user via development configuration",
                    exc_info=True,
                )
                self.user = self.sign_request(
                    {
                        "last_updated": int(time.time()),
                        "username": user_email,
                        "requests": [],
                    }
                )
                self.affected_user = self.user
            else:
                log.error("Unable to get Dynamo table.", exc_info=True)
                raise

    def write_resource_cache_data(self, data):
        self.parallel_write_table(
            self.resource_cache_table, data, ["resourceId", "resourceType"]
        )

    async def get_dynamic_config_yaml(self) -> bytes:
        """Retrieve dynamic configuration yaml."""
        return await sync_to_async(self.get_dynamic_config_yaml_sync)()

    def get_dynamic_config_yaml_sync(self) -> bytes:
        """Retrieve dynamic configuration yaml synchronously"""
        c = b""

        try:
            current_config = self.dynamic_config.get_item(Key={"id": "master"})
            if not current_config:
                return c
            compressed_config = current_config.get("Item", {}).get("config", "")
            if not compressed_config:
                return c
            c = zlib.decompress(compressed_config.value)
        except Exception:  # noqa
            sentry_sdk.capture_exception()
        return c

    def get_dynamic_config_dict(self) -> dict:
        """Retrieve dynamic configuration dictionary that can be merged with primary configuration dictionary."""
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:  # if cleanup: 'RuntimeError: There is no current event loop..'
            loop = None
        if loop and loop.is_running():
            current_config_yaml = self.get_dynamic_config_yaml_sync()
        else:
            current_config_yaml = asyncio.run(self.get_dynamic_config_yaml())
        config_d = yaml.safe_load(current_config_yaml)
        return config_d

    async def get_all_api_health_alerts(self) -> list:
        """Return all requests. If a status is specified, only requests with the specified status will be returned.

        :param status:
        :return:
        """
        response: dict = self.api_health_roles_table.scan()
        items = response.get("Items", [])
        while "LastEvaluatedKey" in response:
            response = self.api_health_roles_table.scan(
                ExclusiveStartKey=response["LastEvaluatedKey"]
            )
            items.extend(self._data_from_dynamo_replace(response["Items"]))

        return items

    async def get_api_health_alert_app(self, app_name) -> dict:
        resp: dict = await sync_to_async(self.api_health_roles_table.get_item)(
            Key={"appName": app_name}
        )
        return resp.get("Item", None)

    async def write_api_health_alert_info(self, request, user_email: str):
        """
        Writes a health alert role to the appropriate DynamoDB table
        """
        function: str = (
            f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
        )
        # enrich request
        request["app_create_time"]: int = int(time.time())
        request["updated_by"]: str = user_email
        request["last_updated"]: int = int(time.time())

        try:
            await sync_to_async(self.api_health_roles_table.put_item)(
                Item=self._data_to_dynamo_replace(request)
            )
        except Exception:
            error = {
                "message": "Unable to add new api_health info request",
                "request": request,
                "function": function,
            }
            log.error(error, exc_info=True)
            raise

        return request

    async def update_api_health_alert_info(
        self, request: dict, user_email=None, update_by=None, last_updated=None
    ):
        """
        Update api_health_alert_info by roleName
        """
        function: str = (
            f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
        )

        # enrich request
        if update_by:
            request["updated_by"] = update_by
        else:
            request["updated_by"] = user_email
        if last_updated:
            request["last_updated"] = last_updated
        else:
            request["last_updated"] = int(time.time())

        try:
            await sync_to_async(self.api_health_roles_table.put_item)(
                Item=self._data_to_dynamo_replace(request)
            )
        except Exception as e:
            error: dict = {
                "function": function,
                "message": "Unable to update api_health_info request",
                "request": request,
                "error": str(e),
            }
            log.error(error, exc_info=True)
            raise Exception(error)
        return request

    async def delete_api_health_alert_info(self, app: str) -> None:
        """
        Delete api_health_alert_info by roleName
        """
        function: str = (
            f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
        )

        try:
            await sync_to_async(self.api_health_roles_table.delete_item)(
                Key={"appName": app}
            )
        except Exception:
            error: dict = {
                "function": function,
                "message": "Unable to delete api_health info",
                "app": app,
            }
            log.error(error, exc_info=True)
            raise

    async def write_policy_request(
        self,
        user_email: str,
        justification: str,
        arn: str,
        policy_name: str,
        policy_changes: dict,
        resources: List[str],
        resource_policies: List[Dict],
        request_time: int = None,
        request_uuid=None,
        policy_status="pending",
        cross_account_request: bool = False,
        dry_run: bool = False,
    ):
        """
        Writes a policy request to the appropriate DynamoDB table
        dry_run will create the request format, but won't actually write it
        Sample run:
        write_policy_request(policy_changes)
        """

        request_time = request_time or int(time.time())

        # Craft the new request json
        timestamp = int(time.time())
        request_id = request_uuid or str(uuid.uuid4())
        new_request = {
            "request_id": request_id,
            "arn": arn,
            "status": policy_status,
            "justification": justification,
            "request_time": request_time,
            "updated_by": user_email,
            "last_updated": timestamp,
            "username": user_email,
            "policy_name": policy_name,
            "policy_changes": json.dumps(policy_changes),
            "resources": resources,
            "resource_policies": resource_policies,
            "cross_account_request": cross_account_request,
        }

        if not dry_run:
            try:
                await sync_to_async(self.policy_requests_table.put_item)(
                    Item=self._data_to_dynamo_replace(new_request)
                )
            except Exception as e:
                error = f"Unable to add new policy request: {new_request}: {str(e)}"
                log.error(error, exc_info=True)
                raise Exception(error)
        else:
            log_data = {
                "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
                "request": new_request,
                "message": "Dry run, skipping adding request to dynamo",
            }
            log.debug(log_data)
        return new_request

    async def write_policy_request_v2(self, extended_request: ExtendedRequestModel):
        """
        Writes a policy request v2 to the appropriate DynamoDB table
        Sample run:
        write_policy_request_v2(request)
        """
        new_request = {
            "request_id": extended_request.id,
            "principal": extended_request.principal.dict(),
            "status": extended_request.request_status.value,
            "justification": extended_request.justification,
            "request_time": extended_request.timestamp,
            "last_updated": int(time.time()),
            "version": "2",
            "extended_request": json.loads(extended_request.json()),
            "username": extended_request.requester_email,
        }

        if extended_request.principal.principal_type == "AwsResource":
            new_request["arn"] = extended_request.principal.principal_arn
        elif extended_request.principal.principal_type == "HoneybeeAwsResourceTemplate":
            repository_name = extended_request.principal.repository_name
            resource_identifier = extended_request.principal.resource_identifier
            new_request["arn"] = f"{repository_name}-{resource_identifier}"
        else:
            raise Exception("Invalid principal type")

        log_data = {
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "message": "Writing policy request v2 to Dynamo",
            "request": new_request,
        }
        log.debug(log_data)

        try:
            await sync_to_async(self.policy_requests_table.put_item)(
                Item=self._data_to_dynamo_replace(new_request)
            )
            log_data[
                "message"
            ] = "Successfully finished writing policy request v2 to Dynamo"
            log.debug(log_data)
        except Exception as e:
            log_data["message"] = "Error occurred writing policy request v2 to Dynamo"
            log_data["error"] = str(e)
            log.error(log_data, exc_info=True)
            error = f"{log_data['message']}: {str(e)}"
            raise Exception(error)

        return new_request

    async def update_policy_request(self, updated_request):
        """
        Update a policy request by request ID
        Sample run:
        update_policy_request(policy_changes)
        """
        updated_request["last_updated"] = int(time.time())
        try:
            await sync_to_async(self.policy_requests_table.put_item)(
                Item=self._data_to_dynamo_replace(updated_request)
            )
        except Exception as e:
            error = f"Unable to add updated policy request: {updated_request}: {str(e)}"
            log.error(error, exc_info=True)
            raise Exception(error)

        return updated_request

    async def get_policy_requests(self, arn=None, request_id=None):
        """Reads a policy request from the appropriate DynamoDB table"""
        if not arn and not request_id:
            raise Exception("Must pass in ARN or policy request ID")
        if request_id:
            requests = self.policy_requests_table.query(
                KeyConditionExpression="request_id = :ri",
                ExpressionAttributeValues={":ri": request_id},
            )
        else:
            requests = self.policy_requests_table.query(
                KeyConditionExpression="arn = :arn",
                ExpressionAttributeValues={":arn": arn},
            )
        matching_requests = []
        if requests["Items"]:
            items = self._data_from_dynamo_replace(requests["Items"])
            items = await self.convert_policy_requests_to_v3(items)
            matching_requests.extend(items)
        return matching_requests

    async def convert_policy_requests_to_v3(self, requests):
        # Remove this function and calls to this function after a grace period of
        changed = False
        for request in requests:
            if not request.get("version") in ["2"]:
                continue
            if request.get("extended_request") and not request.get("principal"):
                principal_arn = request.pop("arn")
                request["principal"] = {
                    "principal_arn": principal_arn,
                    "principal_type": "AwsResource",
                }
                request["extended_request"]["principal"] = {
                    "principal_arn": principal_arn,
                    "principal_type": "AwsResource",
                }
            request.pop("arn", None)
            changes = (
                request.get("extended_request", {})
                .get("changes", {})
                .get("changes", [])
            )
            for change in changes:
                if not change.get("principal_arn"):
                    continue
                if not change.get("version") in ["2.0", "2", 2]:
                    continue
                change["principal"] = {
                    "principal_arn": change["principal_arn"],
                    "principal_type": "AwsResource",
                }
                change.pop("principal_arn")
                change["version"] = "3.0"
                changed = True
        if changed:
            self.parallel_write_table(self.policy_requests_table, requests)
        return requests

    async def get_all_policy_requests(
        self, status: Optional[str] = "pending"
    ) -> List[Dict[str, Union[int, List[str], str]]]:
        """Return all policy requests. If a status is specified, only requests with the specified status will be
        returned.

        :param status:
        :return:
        """
        requests = await sync_to_async(self.parallel_scan_table)(
            self.policy_requests_table
        )

        requests = await self.convert_policy_requests_to_v3(requests)

        return_value = []
        if status:
            for item in requests:
                if status and item["status"] == status:
                    return_value.append(item)
        else:
            return_value = requests

        return return_value

    async def update_dynamic_config(self, config: str, updated_by: str) -> None:
        """Take a YAML config and writes to DDB (The reason we use YAML instead of JSON is to preserve comments)."""
        # Validate that config loads as yaml, raises exception if not
        yaml.safe_load(config)
        stats.count("update_dynamic_config", tags={"updated_by": updated_by})
        current_config_entry = self.dynamic_config.get_item(Key={"id": "master"})
        if current_config_entry.get("Item"):
            old_config = {
                "id": current_config_entry["Item"]["updated_at"],
                "updated_by": current_config_entry["Item"]["updated_by"],
                "config": current_config_entry["Item"]["config"],
                "updated_at": str(int(time.time())),
            }

            self.dynamic_config.put_item(Item=self._data_to_dynamo_replace(old_config))

        new_config = {
            "id": "master",
            "config": zlib.compress(config.encode()),
            "updated_by": updated_by,
            "updated_at": str(int(time.time())),
        }
        self.dynamic_config.put_item(Item=self._data_to_dynamo_replace(new_config))

    def validate_signature(self, items):
        signature = items.pop("signature")
        if isinstance(signature, Binary):
            signature = signature.value
        json_request = json.dumps(items, sort_keys=True)
        if not crypto.verify(json_request, signature):
            raise Exception(f"Invalid signature for request: {json_request}")

    def sign_request(
        self, user_entry: Dict[str, Union[Decimal, List[str], Binary, str]]
    ) -> Dict[str, Union[Decimal, List[str], str, bytes]]:
        """
        Sign the request and returned request with signature
        :param user_entry:
        :return:
        """
        # Remove old signature if it exists
        user_entry.pop("signature", None)
        user_entry = self._data_from_dynamo_replace(user_entry)
        json_request = json.dumps(user_entry, sort_keys=True, use_decimal=True)
        sig = crypto.sign(json_request)
        user_entry["signature"] = sig
        return user_entry

    async def authenticate_user(self, login_attempt) -> AuthenticationResponse:
        function: str = (
            f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
        )
        log_data = {
            "function": function,
            "user_email": login_attempt.username,
            "after_redirect_uri": login_attempt.after_redirect_uri,
        }
        user_entry = await sync_to_async(self.users_table.query)(
            KeyConditionExpression="username = :un",
            ExpressionAttributeValues={":un": login_attempt.username},
        )
        user = None

        generic_error = ["User doesn't exist, or password is incorrect. "]

        if user_entry and "Items" in user_entry and len(user_entry["Items"]) == 1:
            user = user_entry["Items"][0]
        if not user:
            delay_error = await wait_after_authentication_failure(
                login_attempt.username
            )
            error = f"Unable to find user: {login_attempt.username}"
            log.error({**log_data, "message": error + delay_error})
            return AuthenticationResponse(
                authenticated=False, errors=generic_error + [delay_error]
            )

        if not user.get("password"):
            delay_error = await wait_after_authentication_failure(
                login_attempt.username
            )
            error = "User exists, but doesn't have a password stored in the database"
            log.error({**log_data, "message": error + delay_error})
            return AuthenticationResponse(
                authenticated=False, errors=generic_error + [delay_error]
            )

        password_hash_matches = bcrypt.checkpw(
            login_attempt.password.encode("utf-8"), user["password"].value
        )
        if not password_hash_matches:
            delay_error = await wait_after_authentication_failure(
                login_attempt.username
            )
            error = "Password does not match. "
            log.error({**log_data, "message": error + delay_error})
            return AuthenticationResponse(
                authenticated=False, errors=generic_error + [delay_error]
            )
        return AuthenticationResponse(
            authenticated=True, username=user["username"], groups=user["groups"]
        )

    def create_user(
        self,
        user_email,
        password: Optional[str] = None,
        groups: Optional[List[str]] = None,
    ):
        if not groups:
            groups = []
        timestamp = int(time.time())
        unsigned_user_entry = {
            "created": timestamp,
            "last_updated": timestamp,
            "username": user_email,
            "requests": [],
            "groups": groups,
        }

        if password:
            pw = bytes(password, "utf-8")
            salt = bcrypt.gensalt()
            unsigned_user_entry["password"] = bcrypt.hashpw(pw, salt)

        user_entry = self.sign_request(unsigned_user_entry)
        try:
            self.users_table.put_item(Item=self._data_to_dynamo_replace(user_entry))
        except Exception as e:
            error = f"Unable to add user submission: {user_entry}: {str(e)}"
            log.error(error, exc_info=True)
            raise Exception(error)
        return user_entry

    def update_user(
        self,
        user_email,
        password: Optional[str] = None,
        groups: Optional[List[str]] = None,
    ):
        if not groups:
            groups = []

        user_ddb = self.users_table.query(
            KeyConditionExpression="username = :un",
            ExpressionAttributeValues={":un": user_email},
        )

        user = None

        if user_ddb and "Items" in user_ddb and len(user_ddb["Items"]) == 1:
            user = user_ddb["Items"][0]

        if not user:
            raise DataNotRetrievable(f"Unable to find user: {user_email}")

        timestamp = int(time.time())

        if password:
            pw = bytes(password, "utf-8")
            salt = bcrypt.gensalt()
            user["password"] = bcrypt.hashpw(pw, salt)

        if groups:
            user["groups"] = groups
        user["last_updated"] = timestamp

        user_entry = self.sign_request(user)
        try:
            self.users_table.put_item(Item=self._data_to_dynamo_replace(user_entry))
        except Exception as e:
            error = f"Unable to add user submission: {user_entry}: {str(e)}"
            log.error(error, exc_info=True)
            raise Exception(error)
        return user_entry

    def delete_user(self, user_email):
        function: str = (
            f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
        )
        log_data = {"function": function, "user_email": user_email}
        log.debug(log_data)
        user_entry = {"username": user_email}
        try:
            self.users_table.delete_item(Key=self._data_to_dynamo_replace(user_entry))
        except Exception as e:
            error = f"Unable to delete user: {user_entry}: {str(e)}"
            log.error(error, exc_info=True)
            raise Exception(error)

    async def get_user(
        self, user_email: str
    ) -> Optional[Dict[str, Union[Decimal, List[str], Binary, str]]]:
        function: str = (
            f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
        )
        log_data = {"function": function, "user_email": user_email}

        log.debug(log_data)

        user = self.users_table.query(
            KeyConditionExpression="username = :un",
            ExpressionAttributeValues={":un": user_email},
        )

        if user and "Items" in user and len(user["Items"]) == 1:
            return user["Items"][0]
        return None

    def get_or_create_user(
        self, user_email: str
    ) -> Dict[str, Union[Decimal, List[str], Binary, str]]:
        function: str = (
            f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}"
        )
        log_data = {"function": function, "user_email": user_email}

        log.debug(log_data)

        user = self.users_table.query(
            KeyConditionExpression="username = :un",
            ExpressionAttributeValues={":un": user_email},
        )

        items = []

        if user and "Items" in user:
            items = user["Items"]

        if not items:
            return self.create_user(user_email)
        return items[0]

    def resolve_request_ids(
        self, request_ids: List[str]
    ) -> List[Dict[str, Union[int, str]]]:
        requests = []
        for request_id in request_ids:
            request = self.requests_table.query(
                KeyConditionExpression="request_id = :ri",
                ExpressionAttributeValues={":ri": request_id},
            )

            if request["Items"]:
                items = self._data_from_dynamo_replace(request["Items"])
                requests.append(items[0])
            else:
                raise NoMatchingRequest(
                    f"No matching request for request_id: {request_id}"
                )
        return requests

    def add_request_id_to_user(
        self,
        affected_user: Dict[str, Union[Decimal, List[str], Binary, str]],
        request_id: str,
    ) -> None:
        affected_user["requests"].append(request_id)
        self.users_table.put_item(
            Item=self._data_to_dynamo_replace(self.sign_request(affected_user))
        )

    def add_request(
        self,
        user_email: str,
        group: str,
        justification: str,
        request_time: None = None,
        status: str = "pending",
        updated_by: Optional[str] = None,
    ) -> Dict[str, Union[int, str]]:
        """
        Add a user request to the dynamo table

        Sample run:
        add_request("user@example.com", "engtest", "because")

        :param user_email: Email address of user
        :param group: Name of group user is requesting access to
        :param justification:
        :param request_id:
        :param request_time:
        :param status:
        :param updated_by:
        :return:
        """
        """
        Request:
          group
          justification
          role
          request_time
          approval_time
          updated_by
          approval_reason
          status

        user@example.com:
          requests: []
          last_updated: 1
          signature: xxxx
        #pending: []
        #expired: []
        # How to expire requests if soemeone maliciously deletes content
        # How to query for all approved requests for group X
        # What if we want to send email saying your request is expiring in 7 days? Maybe celery to query all
        # What about concept of request ID? Maybe base64 encoded thing?
        # Need an all-in-one page to show all pending requests, all expired/approved requests

      """
        request_time = request_time or int(time.time())

        stats.count("new_group_request", tags={"user": user_email, "group": group})

        if self.affected_user.get("username") != user_email:
            self.affected_user = self.get_or_create_user(user_email)
        # Get current user. Create if they do not already exist
        # self.user = self.get_or_create_user(user_email)
        # Get current user requests, which will validate existing signature
        # existing_request_ids = self.user["requests"]
        # existing_requests = self.resolve_request_ids(existing_request_ids)
        existing_pending_requests_for_group = self.get_requests_by_user(
            user_email, group, status="pending"
        )

        # Craft the new request json
        timestamp = int(time.time())
        request_id = str(uuid.uuid4())
        new_request = {
            "request_id": request_id,
            "group": group,
            "status": status,
            "justification": justification,
            "request_time": request_time,
            "updated_by": updated_by,
            "last_updated": timestamp,
            "username": user_email,
        }

        # See if user already has an active or pending request for the group
        if existing_pending_requests_for_group:
            for request in existing_pending_requests_for_group:
                raise PendingRequestAlreadyExists(
                    f"Pending request for group: {group} already exists: {request}"
                )
        try:
            self.requests_table.put_item(Item=self._data_to_dynamo_replace(new_request))
        except Exception as e:
            error = {
                "error": f"Unable to add user request: {str(e)}",
                "request": new_request,
            }
            log.error(error, exc_info=True)
            raise Exception(error)

        self.add_request_id_to_user(self.affected_user, request_id)

        return new_request

    async def get_all_requests(self, status=None):
        """Return all requests. If a status is specified, only requests with the specified status will be returned.

        :param status:
        :return:
        """
        items = await sync_to_async(self.parallel_scan_table)(self.requests_table)

        return_value = []
        if status:
            for item in items:
                new_json = []
                for j in item["json"]:
                    if j["status"] == status:
                        new_json.append(j)
                item["json"] = new_json
                if new_json:
                    return_value.append(item)
        else:
            return_value = items

        return return_value

    def get_requests_by_user(
        self,
        user_email: str,
        group: str = None,
        status: str = None,
        use_cache: bool = False,
    ) -> dict:
        """Get requests by user. Group and status can also be specified to filter results.
        :param user_email:
        :param group:
        :param status:
        :return:
        """
        red_key = f"USER_REQUESTS_{user_email}-{group}-{status}"

        if use_cache:
            requests_to_return = red.get(red_key)
            if requests_to_return:
                return json.loads(requests_to_return)

        if self.affected_user.get("username") != user_email:
            self.affected_user = self.get_or_create_user(user_email)
        existing_request_ids = self.affected_user["requests"]
        existing_requests = self.resolve_request_ids(existing_request_ids)
        requests_to_return = []
        if existing_requests:
            for request in existing_requests:
                if group and request["group"] != group:
                    continue
                if status and request["status"] != status:
                    continue
                requests_to_return.append(request)
        if use_cache:
            red.setex(red_key, 120, json.dumps(requests_to_return))
        return requests_to_return

    def change_request_status(
        self, user_email, group, new_status, updated_by=None, reviewer_comments=None
    ):
        """

        :param user:
        :param status:
        :param request_id:
        :return:
        """
        stats.count(
            "update_group_request",
            tags={
                "user": user_email,
                "group": group,
                "new_status": new_status,
                "updated_by": updated_by,
            },
        )
        modified_request = None
        if self.affected_user.get("username") != user_email:
            self.affected_user = self.get_or_create_user(user_email)
        timestamp = int(time.time())
        if new_status not in POSSIBLE_STATUSES:
            raise Exception(
                f"Invalid status. Status must be one of {POSSIBLE_STATUSES}"
            )
        if new_status == "approved" and not updated_by:
            raise Exception(
                "You must provide `updated_by` to change a request status to approved."
            )
        existing_requests = self.get_requests_by_user(user_email)
        if existing_requests:
            updated = False
            for request in existing_requests:
                if request["group"] == group:
                    request["updated_by"] = updated_by
                    request["status"] = new_status
                    request["last_updated"] = timestamp
                    request["reviewer_comments"] = reviewer_comments
                    modified_request = request
                    try:
                        self.requests_table.put_item(
                            Item=self._data_to_dynamo_replace(request)
                        )
                    except Exception as e:
                        error = f"Unable to add user request: {request}: {str(e)}"
                        log.error(error, exc_info=True)
                        raise Exception(error)
                    updated = True

            if not updated:
                raise NoExistingRequest(
                    f"Unable to find existing request for user: {user_email} and group: {group}."
                )
        else:
            raise NoExistingRequest(
                f"Unable to find existing requests for user: {user_email}"
            )

        return modified_request

    def change_request_status_by_id(
        self,
        request_id: str,
        new_status: str,
        updated_by: Optional[str] = None,
        reviewer_comments: Optional[str] = None,
    ) -> Dict[str, Union[int, str]]:
        """
        Change request status by ID

        :param request_id:
        :param new_status:
        :param updated_by:
        :return: new requests
        """
        modified_request = None
        if new_status == "approved" and not updated_by:
            raise Exception(
                "You must provide `updated_by` to change a request status to approved."
            )
        requests = self.resolve_request_ids([request_id])

        if new_status not in POSSIBLE_STATUSES:
            raise Exception(
                f"Invalid status. Status must be one of {POSSIBLE_STATUSES}"
            )

        for request in requests:
            request["status"] = new_status
            request["updated_by"] = updated_by
            request["last_updated"] = int(time.time())
            request["reviewer_comments"] = reviewer_comments
            modified_request = request
            try:
                self.requests_table.put_item(Item=self._data_to_dynamo_replace(request))
            except Exception as e:
                error = f"Unable to add user request: {request} : {str(e)}"
                log.error(error, exc_info=True)
                raise Exception(error)
        return modified_request

    def get_all_policies(self):
        """Return all policies."""
        response = self.policies_table.scan()
        items = []

        if response and "Items" in response:
            items = self._data_from_dynamo_replace(response["Items"])

        while "LastEvaluatedKey" in response:
            response = self.policies_table.scan(
                ExclusiveStartKey=response["LastEvaluatedKey"]
            )
            items.extend(self._data_from_dynamo_replace(response["Items"]))

        return items

    async def create_group_log_entry(
        self,
        group: str,
        username: str,
        updated_by: str,
        action: str,
        updated_at: None = None,
        extra: None = None,
    ) -> None:
        updated_at = updated_at or int(time.time())

        log_id = str(uuid.uuid4())
        log_entry = {
            "uuid": log_id,
            "group": group,
            "username": username,
            "updated_by": updated_by,
            "updated_at": updated_at,
            "action": action,
            "extra": extra,
        }
        self.group_log.put_item(Item=self._data_to_dynamo_replace(log_entry))

    async def get_all_audit_logs(self) -> List[Dict[str, Union[int, None, str]]]:
        response = await sync_to_async(self.group_log.scan)()
        items = []

        if response and "Items" in response:
            items = self._data_from_dynamo_replace(response["Items"])

        while "LastEvaluatedKey" in response:
            response = await sync_to_async(self.group_log.scan)(
                ExclusiveStartKey=response["LastEvaluatedKey"]
            )
            items.extend(self._data_from_dynamo_replace(response["Items"]))

        return items

    async def get_all_pending_requests(self):
        return await self.get_all_requests(status="pending")

    def batch_write_cloudtrail_events(self, items):
        with self.cloudtrail_table.batch_writer(
            overwrite_by_pkeys=["arn", "request_id"]
        ) as batch:
            for item in items:
                batch.put_item(Item=self._data_to_dynamo_replace(item))
        return True

    async def get_top_cloudtrail_errors_by_arn(self, arn, n=5):
        response: dict = await sync_to_async(self.cloudtrail_table.query)(
            KeyConditionExpression=Key("arn").eq(arn)
        )
        items = response.get("Items", [])
        aggregated_errors = defaultdict(dict)

        for item in items:
            if int(item["ttl"]) < int(time.time()):
                continue
            event_call = item["event_call"]
            event_resource = item.get("resource", "")

            event_string = f"{event_call}|||{event_resource}"
            if not aggregated_errors.get(event_string):
                aggregated_errors[event_string]["count"] = 0
                aggregated_errors[event_string]["generated_policy"] = item.get(
                    "generated_policy", {}
                )
            aggregated_errors[event_string]["count"] += 1

        top_n_errors = {
            k: v
            for k, v in sorted(
                aggregated_errors.items(),
                key=lambda item: item[1]["count"],
                reverse=True,
            )[:n]
        }

        return top_n_errors

    def count_arn_errors(self, error_count, items):
        for item in items:
            arn = item.get("arn")
            if not error_count.get(arn):
                error_count[arn] = 0
            error_count[arn] += item.get("count", 1)
        return error_count


class IAMRoleDynamoHandler(BaseDynamoHandler):
    def __init__(self) -> None:
        try:
            self.role_table = self._get_dynamo_table(
                config.get("aws.iamroles_dynamo_table", "consoleme_iamroles_global")
            )

        except Exception:
            log.error("Unable to get the IAM Role DynamoDB tables.", exc_info=True)
            raise

    @retry(
        stop_max_attempt_number=4,
        wait_exponential_multiplier=1000,
        wait_exponential_max=1000,
    )
    def _update_role_table_value(self, role_ddb: dict) -> None:
        """Run the specific DynamoDB update with retryability."""
        self.role_table.put_item(Item=role_ddb)

    @retry(
        stop_max_attempt_number=4,
        wait_exponential_multiplier=1000,
        wait_exponential_max=1000,
    )
    def fetch_iam_role(self, role_arn: str, account_id: str):
        return self.role_table.get_item(Key={"arn": role_arn, "accountId": account_id})

    def convert_iam_resource_to_json(self, role: dict) -> str:
        return json.dumps(role, default=self._json_encode_timestamps)

    def _json_encode_timestamps(self, field: datetime) -> str:
        """Solve those pesky timestamps and JSON annoyances."""
        if isinstance(field, datetime):
            return get_iso_string(field)

    def sync_iam_role_for_account(self, role_ddb: dict) -> None:
        """Sync the IAM roles received to DynamoDB.

        :param role_ddb:
        :return:
        """
        try:
            # Unfortunately, DDB does not support batch updates :(... So, we need to update each item individually :/
            self._update_role_table_value(role_ddb)

        except Exception as e:
            log_data = {
                "message": "Error syncing Account's IAM roles to DynamoDB",
                "account_id": role_ddb["accountId"],
                "role_ddb": role_ddb,
                "error": str(e),
            }
            log.error(log_data, exc_info=True)
            raise

    def fetch_all_roles(self):
        response = self.role_table.scan()
        items = []

        if response and "Items" in response:
            items = self._data_from_dynamo_replace(response["Items"])
        while "LastEvaluatedKey" in response:
            response = self.role_table.scan(
                ExclusiveStartKey=response["LastEvaluatedKey"]
            )
            items.extend(self._data_from_dynamo_replace(response["Items"]))
        return items



import tornado.web

from consoleme.config import config
from consoleme.handlers.base import BaseHandler
from consoleme.lib.loader import WebpackLoader
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.models import DataTableResponse

log = config.get_logger()
aws = get_plugin_by_name(config.get("plugins.aws", "default_aws"))()
auth = get_plugin_by_name(config.get("plugins.auth", "default_auth"))()
stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()


# TODO, move followings to util file
async def _filter_by_extension(bundle, extension):
    """Return only files with the given extension"""
    for chunk in bundle:
        if chunk["name"].endswith(".{0}".format(extension)):
            yield chunk


async def _get_bundle(name, extension, config):
    loader = WebpackLoader(name=name, config=config)
    bundle = loader.get_bundle(name)
    if extension:
        bundle = await _filter_by_extension(bundle, extension)
    return bundle


async def get_as_tags(name="main", extension=None, config=config, attrs=""):
    """
    Get a list of formatted <script> & <link> tags for the assets in the
    named bundle.

    :param bundle_name: The name of the bundle
    :param extension: (optional) filter by extension, eg. 'js' or 'css'
    :param config: (optional) the name of the configuration
    :return: a list of formatted tags as strings
    """

    bundle = await _get_bundle(name, extension, config)
    tags = []
    for chunk in bundle:
        if chunk["name"].endswith((".js", ".js.gz")):
            tags.append(
                ('<script type="text/javascript" src="{0}" {1}></script>').format(
                    chunk["url"], attrs
                )
            )
        elif chunk["name"].endswith((".css", ".css.gz")):
            tags.append(
                ('<link type="text/css" href="{0}" rel="stylesheet" {1}/>').format(
                    chunk["url"], attrs
                )
            )
    return tags


class EligibleRoleHandler(BaseHandler):
    async def post(self):
        """
        Post to the index endpoint. This will generate a list of roles the user is eligible to access on the console
        ---
        description: Retrieves a user's eligible roles for AWS console access.
        responses:
            200:
                description: json list of roles
        """

        roles = []
        for arn in self.eligible_roles:
            role_name = arn.split("/")[-1]
            account_id = arn.split(":")[4]
            account_name = self.eligible_accounts.get(account_id, "")
            formatted_account_name = config.get(
                "role_select_page.formatted_account_name", "{account_name}"
            ).format(account_name=account_name, account_id=account_id)
            roles.append(
                {
                    "arn": arn,
                    "account_name": formatted_account_name,
                    "account_id": account_id,
                    "role_name": f"[{role_name}](/policies/edit/{account_id}/iamrole/{role_name})",
                    "redirect_uri": f"/role/{arn}",
                }
            )

        # Default sort by account name
        roles = sorted(roles, key=lambda i: i.get("account_name", 0))
        total_count = len(roles)

        self.set_header("Content-Type", "application/json")
        res = DataTableResponse(
            totalCount=total_count, filteredCount=total_count, data=roles
        )
        self.write(res.json())
        await self.finish()


class EligibleRolePageConfigHandler(BaseHandler):
    async def get(self):
        """
        /eligible_roles_page_config
        ---
        get:
            description: Retrieve Role Page Configuration
            responses:
                200:
                    description: Returns Role Page Configuration
        """
        page_configuration = {
            "pageName": "Select a Role",
            "pageDescription": config.get(
                "role_select_page.table_description",
                "Select a role to login to the AWS console.",
            ),
            "tableConfig": {
                "expandableRows": True,
                "dataEndpoint": "/api/v2/eligible_roles",
                "sortable": False,
                "totalRows": 1000,
                "rowsPerPage": 50,
                "serverSideFiltering": False,
                "allowCsvExport": False,
                "allowJsonExport": False,
                "columns": [
                    {
                        "placeholder": "AWS Console Sign-In",
                        "key": "redirect_uri",
                        "type": "button",
                        "icon": "sign-in",
                        "content": "Sign-In",
                        "onClick": {"action": "redirect"},
                        "style": {"maxWidth": "300px"},
                    },
                    {
                        "placeholder": "Account Name",
                        "key": "account_name",
                        "type": "input",
                    },
                    {"placeholder": "Role Name", "key": "role_name", "type": "link"},
                    {"placeholder": "Account ID", "key": "account_id", "type": "input"},
                ],
            },
        }

        table_configuration = config.get(
            "role_table_config.table_configuration_override", page_configuration
        )

        self.write(table_configuration)


class FrontendHandler(tornado.web.StaticFileHandler):
    def validate_absolute_path(self, root, absolute_path):
        try:
            return super().validate_absolute_path(root, absolute_path)
        except tornado.web.HTTPError as exc:
            if exc.status_code == 404 and self.default_filename is not None:
                absolute_path = self.get_absolute_path(self.root, self.default_filename)
                return super().validate_absolute_path(root, absolute_path)
            raise exc



import json

import jwt
from tornado.testing import AsyncHTTPTestCase
from tornado.web import Application

TEST_SECRET = "SECRET"
TEST_ALG = ["HS256"]


class TestBaseJSONHandler(AsyncHTTPTestCase):
    def get_app(self):
        self.app = Application(self.get_handlers())
        return self.app

    def get_handlers(self):
        from consoleme.handlers.base import BaseJSONHandler

        class JSONHandlerExample(BaseJSONHandler):
            def __init__(self, *args, **kwargs):
                from consoleme.lib.auth import mk_jwt_validator

                TEST_VALIDATOR = mk_jwt_validator(
                    TEST_SECRET, {"alg": {"enum": TEST_ALG}}, {}
                )
                kwargs["jwt_validator"] = TEST_VALIDATOR
                super().__init__(*args, **kwargs)

            def get_app(self):
                self.app = Application(self.get_handlers(), **self.get_app_kwargs())
                return self.app

            def get(self):
                self.write("hello")

        class JSONHandlerMethodOverride(BaseJSONHandler):
            allowed_methods = ["GET", "PUT"]

            def __init__(self, *args, **kwargs):
                from consoleme.lib.auth import mk_jwt_validator

                TEST_VALIDATOR = mk_jwt_validator(
                    TEST_SECRET, {"alg": {"enum": TEST_ALG}}, {}
                )
                kwargs["jwt_validator"] = TEST_VALIDATOR
                super().__init__(*args, **kwargs)

            def get_app(self):
                self.app = Application(self.get_handlers(), **self.get_app_kwargs())
                return self.app

            def get(self):
                self.write("hello")

        return [
            ("/", JSONHandlerExample),
            ("/methodOverride", JSONHandlerMethodOverride),
        ]

    def test_options_default(self):
        response = self.fetch("/", method="OPTIONS")
        self.assertEqual(
            response.headers.get("Access-Control-Allow-Methods"),
            "GET,HEAD,PUT,PATCH,POST,DELETE",
        )

    def test_options_override(self):
        response = self.fetch("/methodOverride", method="OPTIONS")
        self.assertEqual(
            response.headers.get("Access-Control-Allow-Methods"), "GET,PUT"
        )

    def test_missing_auth_header(self):
        response = self.fetch("/")
        self.assertEqual(response.code, 401)

    def test_invalid_auth_header(self):
        response = self.fetch("/", headers={"Authorization": "foo"})
        self.assertEqual(response.code, 401)
        self.assertEqual(json.loads(response.body)["title"], "Unauthorized")
        self.assertEqual(json.loads(response.body)["message"], "Invalid Token")

    def test_invalid_jwt(self):
        payload = {"foo": "bar"}
        tkn = jwt.encode(payload, "WRONG_SECRET", algorithm=TEST_ALG[0])
        response = self.fetch("/", headers={"Authorization": tkn})
        self.assertEqual(response.code, 401)
        self.assertEqual(json.loads(response.body)["title"], "Unauthorized")
        self.assertEqual(
            json.loads(response.body)["message"], "Invalid Token Signature"
        )

    def test_valid_jwt(self):
        payload = {"email": "cnorris@example.com"}
        tkn = jwt.encode(payload, TEST_SECRET, algorithm=TEST_ALG[0])
        response = self.fetch("/", headers={"Authorization": tkn})
        self.assertEqual(response.code, 200)
        self.assertEqual(response.body, b"hello")



import json

from asgiref.sync import async_to_sync
from tornado.testing import AsyncHTTPTestCase


class TestUserRegistrationApi(AsyncHTTPTestCase):
    def get_app(self):
        from consoleme.config import config

        config.CONFIG.config["auth"]["get_user_by_password"] = True
        config.CONFIG.config["auth"]["allow_user_registration"] = True
        from consoleme.routes import make_app

        return make_app(jwt_validator=lambda x: {})

    def tearDown(self) -> None:
        from consoleme.config import config

        config.CONFIG.config["auth"]["get_user_by_password"] = False
        config.CONFIG.config["auth"]["allow_user_registration"] = False

    def test_register_user(self):
        body = json.dumps(
            {
                "username": "testuser5",
                "password": "testuser5password",
            }
        )
        response = self.fetch("/api/v2/user_registration", method="POST", body=body)
        self.assertEqual(response.code, 403)
        self.assertEqual(
            json.loads(response.body),
            {
                "status": "error",
                "reason": "invalid_request",
                "status_code": 403,
                "errors": [
                    "The email address is not valid. It must have exactly one @-sign."
                ],
            },
        )

        body = json.dumps(
            {
                "username": "testuser5@example.com",
                "password": "testuser5password",
            }
        )
        response = self.fetch("/api/v2/user_registration", method="POST", body=body)
        self.assertEqual(response.code, 200)
        self.assertEqual(
            json.loads(response.body),
            {
                "status": "success",
                "status_code": 200,
                "message": "Successfully created user testuser5@example.com.",
            },
        )


class TestLoginApi(AsyncHTTPTestCase):
    def get_app(self):
        from consoleme.config import config

        config.CONFIG.config["auth"]["get_user_by_password"] = True
        config.CONFIG.config["auth"]["set_auth_cookie"] = True
        from consoleme.routes import make_app

        return make_app(jwt_validator=lambda x: {})

    def tearDown(self) -> None:
        from consoleme.config import config

        config.CONFIG.config["auth"]["get_user_by_password"] = False
        config.CONFIG.config["auth"]["set_auth_cookie"] = False

    def test_login_post_no_user(self):
        body = json.dumps(
            {"username": "fakeuser", "password": "pass", "after_redirect_uri": "/"}
        )
        response = self.fetch("/api/v2/login", method="POST", body=body)
        self.assertEqual(response.code, 403)
        self.assertEqual(
            json.loads(response.body),
            {
                "status": "error",
                "reason": "authentication_failure",
                "status_code": 403,
                "errors": [
                    "User doesn't exist, or password is incorrect. ",
                    "Your next authentication failure will result in a 1 second wait. This wait time will expire after 60 seconds of no authentication failures.",
                ],
            },
        )

    def test_login_post_invalid_password(self):
        from consoleme.lib.dynamo import UserDynamoHandler

        ddb = UserDynamoHandler()
        ddb.create_user("testuser", "correctpassword", ["group1", "group2@example.com"])
        body = json.dumps(
            {"username": "testuser", "password": "wrongpass", "after_redirect_uri": "/"}
        )
        response = self.fetch("/api/v2/login", method="POST", body=body)
        self.assertEqual(response.code, 403)
        self.assertEqual(
            json.loads(response.body),
            {
                "status": "error",
                "reason": "authentication_failure",
                "status_code": 403,
                "errors": [
                    "User doesn't exist, or password is incorrect. ",
                    "Your next authentication failure will result in a 1 second wait. This wait time will expire after 60 seconds of no authentication failures.",
                ],
            },
        )

    def test_login_post_success(self):
        from consoleme.lib.dynamo import UserDynamoHandler

        ddb = UserDynamoHandler()
        ddb.create_user(
            "testuser2", "correctpassword", ["group1", "group2@example.com"]
        )
        body = json.dumps(
            {
                "username": "testuser2",
                "password": "correctpassword",
                "after_redirect_uri": "/",
            }
        )
        response = self.fetch("/api/v2/login", method="POST", body=body)
        self.assertEqual(response.code, 200)
        self.assertEqual(
            json.loads(response.body),
            {
                "status": "redirect",
                "reason": "authenticated_redirect",
                "redirect_url": "/",
                "status_code": 200,
                "message": "User has successfully authenticated. Redirecting to their intended destination.",
            },
        )


class TestUserApi(AsyncHTTPTestCase):
    def get_app(self):
        from consoleme.routes import make_app

        return make_app(jwt_validator=lambda x: {})

    def test_create_user(self):
        from consoleme.config import config

        headers = {
            config.get("auth.user_header_name"): "consoleme_admins@example.com",
            config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }

        body = json.dumps(
            {
                "user_management_action": "create",
                "username": "testuser3",
                "password": "testuser3password",
                "groups": ["group1", "group2", "group3"],
            }
        )
        response = self.fetch("/api/v2/user", headers=headers, method="POST", body=body)
        self.assertEqual(response.code, 200)
        self.assertEqual(
            json.loads(response.body),
            {
                "status": "success",
                "status_code": 200,
                "message": "Successfully created user testuser3.",
            },
        )

        # Verify new user works
        from consoleme.lib.dynamo import UserDynamoHandler
        from consoleme.models import LoginAttemptModel

        ddb = UserDynamoHandler()
        login_attempt_success = LoginAttemptModel(
            username="testuser3", password="testuser3password"
        )

        should_pass = async_to_sync(ddb.authenticate_user)(login_attempt_success)
        self.assertEqual(
            should_pass.dict(),
            {
                "authenticated": True,
                "errors": None,
                "username": "testuser3",
                "groups": ["group1", "group2", "group3"],
            },
        )
        login_attempt_fail = LoginAttemptModel(
            username="testuser3", password="wrongpassword"
        )

        should_fail = async_to_sync(ddb.authenticate_user)(login_attempt_fail)

        self.assertEqual(
            should_fail.dict(),
            {
                "authenticated": False,
                "errors": [
                    "User doesn't exist, or password is incorrect. ",
                    "Your next authentication failure will result in a 1 second wait. "
                    "This wait time will expire after 60 seconds of no authentication failures.",
                ],
                "groups": None,
                "username": None,
            },
        )

        # Update password
        body = json.dumps(
            {
                "user_management_action": "update",
                "username": "testuser3",
                "password": "testuser3newpassword",
            }
        )
        response = self.fetch("/api/v2/user", headers=headers, method="POST", body=body)
        self.assertEqual(response.code, 200)
        self.assertEqual(
            json.loads(response.body),
            {
                "status": "success",
                "status_code": 200,
                "message": "Successfully updated user testuser3.",
            },
        )

        # Update groups
        body = json.dumps(
            {
                "user_management_action": "update",
                "username": "testuser3",
                "groups": ["group1", "group2", "group3", "newgroup"],
            }
        )
        response = self.fetch("/api/v2/user", headers=headers, method="POST", body=body)
        self.assertEqual(response.code, 200)
        self.assertEqual(
            json.loads(response.body),
            {
                "status": "success",
                "status_code": 200,
                "message": "Successfully updated user testuser3.",
            },
        )
        # Update groups and password AT THE SAME TIME!!1
        body = json.dumps(
            {
                "user_management_action": "update",
                "username": "testuser3",
                "password": "testuser3newpassword2",
                "groups": ["group1", "group2", "group3", "newgroup", "newgroup2"],
            }
        )
        response = self.fetch("/api/v2/user", headers=headers, method="POST", body=body)
        self.assertEqual(response.code, 200)
        self.assertEqual(
            json.loads(response.body),
            {
                "status": "success",
                "status_code": 200,
                "message": "Successfully updated user testuser3.",
            },
        )

        # Delete the user

        body = json.dumps(
            {
                "user_management_action": "delete",
                "username": "testuser3",
            }
        )

        response = self.fetch("/api/v2/user", headers=headers, method="POST", body=body)
        self.assertEqual(response.code, 200)
        self.assertEqual(
            json.loads(response.body),
            {
                "status": "success",
                "status_code": 200,
                "message": "Successfully deleted user testuser3.",
            },
        )



import sentry_sdk
import tornado.escape
import ujson as json
from asgiref.sync import sync_to_async
from cloudaux.aws.iam import (
    get_managed_policy_document,
    get_role_managed_policy_documents,
    get_user_managed_policy_documents,
)

from consoleme.config import config
from consoleme.exceptions.exceptions import MustBeFte
from consoleme.handlers.base import BaseAPIV2Handler
from consoleme.lib.aws import get_all_iam_managed_policies_for_account
from consoleme.models import Status2, WebResponse

log = config.get_logger()


class ManagedPoliciesOnPrincipalHandler(BaseAPIV2Handler):
    """
    Handler for /api/v2/managed_policies_on_principal/{arn}

    Returns managed policy and latest policy version information for a principal
    """

    async def get(self, arn):
        if config.get("policy_editor.disallow_contractors", True) and self.contractor:
            if self.user not in config.get(
                "groups.can_bypass_contractor_restrictions", []
            ):
                raise MustBeFte("Only FTEs are authorized to view this page.")

        errors = []
        if not arn.startswith("arn:aws:iam::"):
            errors.append("ARN must start with 'arn:aws:iam::'")

        principal_name = tornado.escape.xhtml_escape(arn.split("/")[-1])
        principal_type = tornado.escape.xhtml_escape(arn.split(":")[5].split("/")[0])
        account_id = tornado.escape.xhtml_escape(arn.split(":")[4])
        if principal_type not in ["role", "user"]:
            errors.append(f"Principal type must be role or user. not {principal_type}")

        log_data = {
            "function": "ManagedPoliciesOnRoleHandler.get",
            "user": self.user,
            "ip": self.ip,
            "message": "Retrieving managed policies for role",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
            "account_id": account_id,
            "principal_name": principal_name,
            "principal_type": principal_type,
        }

        log.debug(log_data)
        if errors:
            log.error(
                {**log_data, "errors": errors, "message": "Unable to process request"}
            )
            res = WebResponse(
                status=Status2.error,
                reason="bad_request",
                status_code=400,
                errors=errors,
            )
            self.write(res.json())
            return

        if principal_type == "role":
            managed_policy_details = await sync_to_async(
                get_role_managed_policy_documents
            )(
                {"RoleName": principal_name},
                account_number=account_id,
                assume_role=config.get("policies.role_name"),
                region=config.region,
                retry_max_attempts=2,
                client_kwargs=config.get("boto3.client_kwargs", {}),
            )
        elif principal_type == "user":
            managed_policy_details = await sync_to_async(
                get_user_managed_policy_documents
            )(
                {"UserName": principal_name},
                account_number=account_id,
                assume_role=config.get("policies.role_name"),
                region=config.region,
                retry_max_attempts=2,
                client_kwargs=config.get("boto3.client_kwargs", {}),
            )
        else:
            raise Exception("Invalid principal type")
        res = WebResponse(
            status=Status2.success,
            status_code=200,
            data=managed_policy_details,
        )
        self.write(res.json())


class ManagedPoliciesHandler(BaseAPIV2Handler):
    """
    Handler for /api/v2/managed_policies/{policyArn}

    Returns details about the specified managed policy
    """

    async def get(self, policy_arn: str):
        if config.get("policy_editor.disallow_contractors", True) and self.contractor:
            if self.user not in config.get(
                "groups.can_bypass_contractor_restrictions", []
            ):
                raise MustBeFte("Only FTEs are authorized to view this page.")

        account_id = policy_arn.split(":")[4]
        policy_name = policy_arn.split("/")[-1]
        log_data = {
            "function": "ManagedPoliciesHandler.get",
            "user": self.user,
            "ip": self.ip,
            "message": "Retrieving managed policy",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
            "account_id": account_id,
            "policy_name": policy_name,
            "policy_arn": policy_arn,
        }

        log.debug(log_data)

        managed_policy_details = await sync_to_async(get_managed_policy_document)(
            policy_arn=policy_arn,
            account_number=account_id,
            assume_role=config.get("policies.role_name"),
            region=config.region,
            retry_max_attempts=2,
            client_kwargs=config.get("boto3.client_kwargs", {}),
        )
        res = WebResponse(
            status=Status2.success,
            status_code=200,
            data=managed_policy_details,
        )
        self.write(res.json())


class ManagedPoliciesForAccountHandler(BaseAPIV2Handler):
    async def get(self, account_id):
        """
        Retrieve a list of managed policies for an account.
        """
        if config.get("policy_editor.disallow_contractors", True) and self.contractor:
            if self.user not in config.get(
                "groups.can_bypass_contractor_restrictions", []
            ):
                raise MustBeFte("Only FTEs are authorized to view this page.")
        try:
            all_account_managed_policies = (
                await get_all_iam_managed_policies_for_account(account_id)
            )
        except Exception:
            sentry_sdk.capture_exception()
            all_account_managed_policies = []
        self.write(json.dumps(all_account_managed_policies))



import sys

import tornado.escape
import ujson as json
from tornado.httpclient import AsyncHTTPClient, HTTPClientError, HTTPRequest
from tornado.httputil import HTTPHeaders

from consoleme.config import config
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.lib.policies import get_policy_request_uri_v2
from consoleme.models import ExtendedRequestModel

log = config.get_logger()
stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()


async def send_slack_notification_new_request(
    extended_request: ExtendedRequestModel, admin_approved, approval_probe_approved
):
    """
    Sends a notification using specified webhook URL about a new request created
    """
    if not config.get("slack.notifications_enabled", False):
        return

    function = f"{__name__}.{sys._getframe().f_code.co_name}"
    requester = extended_request.requester_email
    arn = extended_request.principal.principal_arn
    stats.count(function, tags={"user": requester, "arn": arn})

    log_data: dict = {
        "function": f"{__name__}.{sys._getframe().f_code.co_name}",
        "user": requester,
        "arn": arn,
        "message": "Incoming request for slack notification",
        "request": extended_request.dict(),
        "admin_approved": admin_approved,
        "approval_probe_approved": approval_probe_approved,
    }
    log.debug(log_data)
    slack_webhook_url = config.get("slack.webhook_url")
    if not slack_webhook_url:
        log_data["message"] = "Missing webhook URL for slack notification"
        log.error(log_data)
        return
    payload = await get_payload(
        extended_request, requester, arn, admin_approved, approval_probe_approved
    )
    http_headers = HTTPHeaders({"Content-Type": "application/json"})
    http_req = HTTPRequest(
        url=slack_webhook_url,
        method="POST",
        headers=http_headers,
        body=json.dumps(payload),
    )
    http_client = AsyncHTTPClient(force_instance=True)
    try:
        await http_client.fetch(request=http_req)
        log_data["message"] = "Slack notification sent"
        log.debug(log_data)
    except (ConnectionError, HTTPClientError) as e:
        log_data["message"] = "Error occurred sending slack notification"
        log_data["error"] = str(e)
        log.error(log_data)


async def get_payload(
    extended_request: ExtendedRequestModel,
    requester: str,
    arn: str,
    admin_approved: bool,
    approval_probe_approved: bool,
):
    request_uri = await get_policy_request_uri_v2(extended_request)
    pre_text = "A new request has been created"
    if admin_approved:
        pre_text += " and auto-approved by admin"
    elif approval_probe_approved:
        pre_text += " and auto-approved by auto-approval probe"

    payload = {
        "blocks": [
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": f"*<{request_uri}|ConsoleMe Policy Change Request>*",
                },
            },
            {
                "type": "section",
                "text": {"type": "mrkdwn", "text": f"*User* \n {requester}"},
            },
            {
                "type": "section",
                "text": {"type": "mrkdwn", "text": f"*Resource* \n {arn}"},
            },
            {
                "type": "section",
                "fields": [
                    {"text": "*Justification*", "type": "mrkdwn"},
                    {"type": "plain_text", "text": "\n"},
                    {
                        "type": "plain_text",
                        "text": f"{tornado.escape.xhtml_escape(extended_request.justification)}",
                    },
                ],
            },
            {"type": "divider"},
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": f"{pre_text}. Click *<{request_uri}|here>* to view it.",
                },
            },
        ]
    }
    return payload



import sys
from typing import Optional

import sentry_sdk
import tornado.escape

from consoleme.config import config
from consoleme.handlers.base import BaseAPIV2Handler
from consoleme.lib.auth import can_delete_iam_principals
from consoleme.lib.aws import delete_iam_user
from consoleme.lib.crypto import Crypto
from consoleme.lib.generic import str2bool
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.lib.v2.aws_principals import get_user_details

stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
log = config.get_logger()
crypto = Crypto()
auth = get_plugin_by_name(config.get("plugins.auth", "default_auth"))()
aws = get_plugin_by_name(config.get("plugins.aws", "default_aws"))()
group_mapping = get_plugin_by_name(
    config.get("plugins.group_mapping", "default_group_mapping")
)()
internal_policies = get_plugin_by_name(
    config.get("plugins.internal_policies", "default_policies")
)()


class UserDetailHandler(BaseAPIV2Handler):
    """Handler for /api/v2/users/{accountNumber}/{userName}

    Allows read and delete access to a specific user in an account.
    """

    allowed_methods = ["GET", "DELETE"]

    def initialize(self):
        self.user: Optional[str] = None
        self.eligible_roles: list = []

    async def get(self, account_id, user_name):
        """
        GET /api/v2/users/{account_number}/{user_name}
        """
        log_data = {
            "function": "UsersDetailHandler.get",
            "user": self.user,
            "ip": self.ip,
            "message": "Retrieving user details",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
            "account_id": account_id,
            "user_name": user_name,
        }
        stats.count(
            "UsersDetailHandler.get",
            tags={"user": self.user, "account_id": account_id, "user_name": user_name},
        )
        log.debug(log_data)
        force_refresh = str2bool(
            self.request.arguments.get("force_refresh", [False])[0]
        )

        error = ""

        try:
            user_details = await get_user_details(
                account_id, user_name, extended=True, force_refresh=force_refresh
            )
        except Exception as e:
            sentry_sdk.capture_exception()
            log.error({**log_data, "error": e}, exc_info=True)
            user_details = None
            error = str(e)

        if not user_details:
            self.send_error(
                404,
                message=f"Unable to retrieve the specified user: {account_id}/{user_name}. {error}",
            )
            return
        self.write(user_details.json())

    async def delete(self, account_id, iam_user_name):
        """
        DELETE /api/v2/users/{account_id}/{iam_user_name}
        """
        account_id = tornado.escape.xhtml_escape(account_id)
        iam_user_name = tornado.escape.xhtml_escape(iam_user_name)

        if not self.user:
            self.write_error(403, message="No user detected")
            return

        log_data = {
            "user": self.user,
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
            "ip": self.ip,
            "account": account_id,
            "iam_user_name": iam_user_name,
        }

        can_delete_principal = can_delete_iam_principals(self.user, self.groups)
        if not can_delete_principal:
            stats.count(
                f"{log_data['function']}.unauthorized",
                tags={
                    "user": self.user,
                    "account": account_id,
                    "iam_user_name": iam_user_name,
                    "authorized": can_delete_principal,
                    "ip": self.ip,
                },
            )
            log_data["message"] = "User is unauthorized to delete an AWS IAM User"
            log.error(log_data)
            self.write_error(
                403, message="User is unauthorized to delete an AWS IAM user"
            )
            return
        try:
            await delete_iam_user(account_id, iam_user_name, self.user)
        except Exception as e:
            log_data["message"] = "Exception deleting AWS IAM User"
            log.error(log_data, exc_info=True)
            stats.count(
                f"{log_data['function']}.exception",
                tags={
                    "user": self.user,
                    "account": account_id,
                    "iam_user_name": iam_user_name,
                    "authorized": can_delete_principal,
                    "ip": self.ip,
                },
            )
            self.write_error(500, message="Error occurred deleting IAM user: " + str(e))
            return

        # if here, user has been successfully deleted
        arn = f"arn:aws:iam::{account_id}:user/{iam_user_name}"
        await aws.fetch_iam_user(account_id, arn)
        response_json = {
            "status": "success",
            "message": "Successfully deleted AWS IAM user from account",
            "iam_user_name": iam_user_name,
            "account": account_id,
        }
        self.write(response_json)



"""
    This script will retrieve and save ConsoleMe's configuration if the specified environment variables are defined.
    This is useful for Docker / ECS.

    * If `CONSOLEME_CONFIG_B64` is defined, this script will decode the Base64 encoded string and save the configuration
    file in the location specified by `CONFIG_LOCATION` (Defaults to: /etc/consoleme/config/config.yaml)

    * If `CONSOLEME_CONFIG_S3` is defined, this script will download the configuration from S3, and store it in the
    location specified by `CONFIG_LOCATION` (Defaults to: /etc/consoleme/config/config.yaml)

    Usage:
        You can Base64 encode ConsoleMe's configuration with the following command:

        export CONSOLEME_CONFIG_B64=`base64 /path/to/your/consoleme/config.yaml`

        Alternatively, you can upload ConsoleMe's configuration to S3 and this script will attempt to download it:

        export CONSOLEME_CONFIG_S3=`s3://location/to/your/config.yaml`

"""

import base64
import errno
import logging
import os
import sys

import boto3

format_c = "%(asctime)s - %(levelname)s - %(name)s - [%(filename)s:%(lineno)s - %(funcName)s() ] - %(message)s"
logging.basicConfig(level=logging.DEBUG, format=format_c)


def split_s3_path(s3_path):
    path_parts = s3_path.replace("s3://", "").split("/")
    b = path_parts.pop(0)
    k = "/".join(path_parts)
    return b, k


def make_directories(loc):
    head, _ = os.path.split(loc)
    try:
        os.makedirs(head)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


save_location = os.environ.get("CONFIG_LOCATION", "/etc/consoleme/config/config.yaml")

base64_config = os.environ.get("CONSOLEME_CONFIG_B64")
s3_config_path = os.environ.get("CONSOLEME_CONFIG_S3")

if not base64_config and not s3_config_path:
    logging.warning(
        "Neither the CONSOLEME_CONFIG_B64 or CONSOLEME_CONFIG_S3 environment variables are defined. This is required "
        "to retrieve a custom ConsoleMe configuration via S3, or decode it from Base64. This warning is safe to ignore "
        "if you're just testing ConsoleMe and want to rely on the default configuration."
    )
    sys.exit(0)
if base64_config and s3_config_path:
    logging.warning(
        "ConsoleMe only expects one environment variable for configuration retrieval to be defined, but found two. "
        "Please only set one of: CONSOLEME_CONFIG_B64 or CONSOLEME_CONFIG_S3."
    )
    sys.exit(1)
if base64_config:
    decoded_config = base64.b64decode(base64_config)
    logging.info(f"Decoded ConsoleMe's configuration and saving to {save_location}")
    make_directories(save_location)
    with open(save_location, "w") as f:
        f.write(decoded_config.decode())
elif s3_config_path:
    client = boto3.client("s3")
    bucket, key = split_s3_path(s3_config_path)
    obj = client.get_object(Bucket=bucket, Key=key)
    s3_object_content = obj["Body"].read()
    logging.info(
        f"Retrieved ConsoleMe's configuration from {s3_config_path} and writing  to {save_location}"
    )
    make_directories(save_location)
    with open(save_location, "w") as f:
        f.write(s3_object_content.decode())



import argparse
import concurrent.futures
import os
import time
from concurrent.futures.thread import ThreadPoolExecutor

from asgiref.sync import async_to_sync
from consoleme_default.plugins.plugins.celery_tasks import (
    celery_tasks as default_celery_tasks,
)

from consoleme.celery_tasks import celery_tasks as celery
from consoleme.config import config
from consoleme.lib.account_indexers import get_account_id_to_name_mapping

start_time = int(time.time())


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


parser = argparse.ArgumentParser(description="Populate ConsoleMe's Redis Cache")
parser.add_argument(
    "--use_celery",
    default=False,
    type=str2bool,
    help="Invoke celery tasks instead of running synchronously",
)
args = parser.parse_args()

# Force dynamic configuration update synchronously
config.CONFIG.load_config_from_dynamo()

if args.use_celery:
    # Initialize Redis locally. If use_celery is set to `True`, you must be running a celery beat and worker. You can
    # run this locally with the following command:
    # `celery -A consoleme.celery_tasks.celery_tasks worker -l DEBUG -B -E --concurrency=8`

    celery.cache_iam_resources_across_accounts()
    celery.cache_s3_buckets_across_accounts()
    celery.cache_sns_topics_across_accounts()
    celery.cache_sqs_queues_across_accounts()
    celery.cache_managed_policies_across_accounts()
    default_celery_tasks.cache_application_information()
    celery.cache_resources_from_aws_config_across_accounts()
    celery.cache_policies_table_details.apply_async(countdown=180)
    celery.cache_policy_requests()
    celery.cache_credential_authorization_mapping.apply_async(countdown=180)

else:
    celery.cache_cloud_account_mapping()
    accounts_d = async_to_sync(get_account_id_to_name_mapping)(force_sync=True)
    default_celery_tasks.cache_application_information()
    executor = ThreadPoolExecutor(max_workers=os.cpu_count())
    futures = []
    for account_id in accounts_d.keys():
        futures.extend(
            [
                executor.submit(celery.cache_iam_resources_for_account, account_id),
                executor.submit(celery.cache_s3_buckets_for_account, account_id),
                executor.submit(celery.cache_sns_topics_for_account, account_id),
                executor.submit(celery.cache_sqs_queues_for_account, account_id),
                executor.submit(celery.cache_managed_policies_for_account, account_id),
                executor.submit(
                    celery.cache_resources_from_aws_config_for_account, account_id
                ),
            ]
        )
    for future in concurrent.futures.as_completed(futures):
        try:
            data = future.result()
        except Exception as exc:
            print("%r generated an exception: %s" % (future, exc))

    celery.cache_policies_table_details()
    celery.cache_policy_requests()
    celery.cache_credential_authorization_mapping()
    # Forces writing config to S3
    celery.cache_iam_resources_across_accounts(
        wait_for_subtask_completion=False, run_subtasks=False
    )
    celery.cache_s3_buckets_across_accounts(
        wait_for_subtask_completion=False, run_subtasks=False
    )
    celery.cache_sns_topics_across_accounts(
        wait_for_subtask_completion=False, run_subtasks=False
    )
    celery.cache_sqs_queues_across_accounts(
        wait_for_subtask_completion=False, run_subtasks=False
    )
    celery.cache_resources_from_aws_config_across_accounts(
        wait_for_subtask_completion=False, run_subtasks=False
    )
    celery.cache_resource_templates_task()
    celery.cache_self_service_typeahead_task()
total_time = int(time.time()) - start_time
print(f"Done caching data in Redis. It took {total_time} seconds")



import asyncio
from typing import Dict, List, Optional, Union

from password_strength import PasswordPolicy

from consoleme.config import config
from consoleme.lib.redis import RedisHandler

red = RedisHandler().redis_sync()


async def wait_after_authentication_failure(user) -> str:
    redix_key_expiration = config.get(
        "wait_after_authentication_failure.expiration", 60
    )
    redis_key = f"wait_after_authentication_failure_{user}"
    num_password_failures = red.get(redis_key)
    if not num_password_failures:
        num_password_failures = 0
    num_password_failures = int(num_password_failures)  # Redis values are strings
    red.setex(redis_key, redix_key_expiration, num_password_failures + 1)
    await asyncio.sleep(num_password_failures ** 2)
    next_delay = (num_password_failures + 1) ** 2
    return (
        f"Your next authentication failure will result in a {next_delay} second wait. "
        f"This wait time will expire after {redix_key_expiration} seconds of no authentication failures."
    )


async def check_password_strength(
    password,
) -> Optional[Union[Dict[str, str], Dict[str, List[str]]]]:
    password_policy_args = {
        "strength": config.get("auth.password_policy.strength", 0.5),
        "entropy_bits": config.get("auth.password_policy.entry_bits"),
        "length": config.get("auth.password_policy.length"),
        "uppercase": config.get("auth.password_policy.uppercase"),
        "numbers": config.get("auth.password_policy.numbers"),
        "special": config.get("auth.password_policy.special"),
        "nonletters": config.get("auth.password_policy.nonletters"),
    }

    # We remove entries with null values since password_strength doesn't like them.
    password_policy_args = {k: v for k, v in password_policy_args.items() if v}

    policy = PasswordPolicy.from_names(**password_policy_args)

    tested_pass = policy.password(password)
    errors = tested_pass.test()
    # Convert errors to string so they can be json encoded later
    errors: List[str] = [str(e) for e in errors]

    if errors:
        return {"message": "Password doesn't have enough entropy.", "errors": errors}



"""
Configuration stack for running ConsoleMe on ECS
"""

from uuid import uuid4

import yaml
from aws_cdk import aws_cognito as cognito
from aws_cdk import aws_elasticache as ec
from aws_cdk import aws_iam as iam
from aws_cdk import aws_lambda as lambda_
from aws_cdk import aws_logs as logs
from aws_cdk import core as cdk
from aws_cdk import custom_resources as cr

from cdk.consoleme_ecs_service.constants import CONFIG_SECRET_NAME
from consoleme_ecs_cdk.service.helpers import create_dependencies_layer


class ConfigStack(cdk.NestedStack):
    """
    Configuration stack for running ConsoleMe on ECS
    """

    def __init__(
        self,
        scope: cdk.Construct,
        id: str,
        cognito_user_pool: cognito.UserPool,
        s3_bucket_name: str,
        create_configuration_lambda_role_arn: str,
        redis: ec.CfnCacheCluster,
        domain_name: str,
        **kwargs
    ) -> None:
        super().__init__(scope, id, **kwargs)

        config_yaml = yaml.load(open("config.yaml"), Loader=yaml.FullLoader)
        spoke_accounts = config_yaml.get("spoke_accounts", [])

        cognito_user_pool_client = cognito.UserPoolClient(
            self,
            "UserPoolClient",
            user_pool=cognito_user_pool,
            generate_secret=True,
            supported_identity_providers=[
                cognito.UserPoolClientIdentityProvider.COGNITO
            ],
            prevent_user_existence_errors=True,
            o_auth=cognito.OAuthSettings(
                callback_urls=[
                    "https://" + domain_name + "/auth",
                    "https://" + domain_name + "/oauth2/idpresponse",
                ],
                logout_urls=["https://" + domain_name + "/logout"],
                flows=cognito.OAuthFlows(
                    authorization_code_grant=True, implicit_code_grant=True
                ),
                scopes=[cognito.OAuthScope.OPENID, cognito.OAuthScope.EMAIL],
            ),
            auth_flows=cognito.AuthFlow(user_password=True, user_srp=True),
        )

        describe_cognito_user_pool_client = cr.AwsCustomResource(
            self,
            "UserPoolClientIDResource",
            policy=cr.AwsCustomResourcePolicy.from_sdk_calls(
                resources=cr.AwsCustomResourcePolicy.ANY_RESOURCE
            ),
            on_create=cr.AwsSdkCall(
                service="CognitoIdentityServiceProvider",
                action="describeUserPoolClient",
                parameters={
                    "UserPoolId": cognito_user_pool.user_pool_id,
                    "ClientId": cognito_user_pool_client.user_pool_client_id,
                },
                physical_resource_id=cr.PhysicalResourceId.of(
                    cognito_user_pool_client.user_pool_client_id
                ),
            ),
            install_latest_aws_sdk=True,
            log_retention=logs.RetentionDays.ONE_WEEK,
        )

        cognito_user_pool_client_secret = (
            describe_cognito_user_pool_client.get_response_field(
                "UserPoolClient.ClientSecret"
            )
        )

        imported_create_configuration_lambda_role = iam.Role.from_role_arn(
            self,
            "ImportedCreateConfigurationFileLambdaRole",
            role_arn=create_configuration_lambda_role_arn,
        )

        jwt_secret = config_yaml["jwt_secret"]

        config_secret_dict = {
            "oidc_secrets": {
                "client_id": cognito_user_pool_client.user_pool_client_id,
                "secret": cognito_user_pool_client_secret,
                "client_scope": ["email", "openid"],
            },
            "jwt_secret": jwt_secret,
        }

        config_secret_yaml = yaml.dump(
            config_secret_dict,
            explicit_start=True,
            default_flow_style=False,
        )

        config_secret = cr.AwsCustomResource(
            self,
            "ConfigSecretResource",
            policy=cr.AwsCustomResourcePolicy.from_sdk_calls(
                resources=cr.AwsCustomResourcePolicy.ANY_RESOURCE
            ),
            on_update=cr.AwsSdkCall(
                service="SecretsManager",
                action="updateSecret",
                parameters={
                    "SecretId": CONFIG_SECRET_NAME,
                    "SecretString": config_secret_yaml,
                },
                physical_resource_id=cr.PhysicalResourceId.from_response("Name"),
            ),
            on_create=cr.AwsSdkCall(
                service="SecretsManager",
                action="createSecret",
                parameters={
                    "Name": CONFIG_SECRET_NAME,
                    "Description": "Sensitive configuration parameters for ConsoleMe",
                    "SecretString": config_secret_yaml,
                },
                physical_resource_id=cr.PhysicalResourceId.from_response("Name"),
            ),
            on_delete=cr.AwsSdkCall(
                service="SecretsManager",
                action="deleteSecret",
                parameters={
                    "SecretId": CONFIG_SECRET_NAME,
                    "ForceDeleteWithoutRecovery": True,
                },
            ),
            install_latest_aws_sdk=True,
            log_retention=logs.RetentionDays.ONE_WEEK,
        )

        create_configuration_lambda = lambda_.Function(
            self,
            "CreateConfigurationFileLambda",
            code=lambda_.Code.from_asset("resources/create_config_lambda"),
            handler="index.handler",
            timeout=cdk.Duration.seconds(30),
            layers=[create_dependencies_layer(self, "create_config_lambda")],
            runtime=lambda_.Runtime.PYTHON_3_8,
            role=imported_create_configuration_lambda_role,
            environment={
                "DEPLOYMENT_BUCKET": s3_bucket_name,
                "OIDC_METADATA_URL": "https://cognito-idp."
                + self.region
                + ".amazonaws.com/"
                + cognito_user_pool.user_pool_id
                + "/.well-known/openid-configuration",
                "REDIS_HOST": redis.attr_redis_endpoint_address,
                "SES_IDENTITY_ARN": "arn:aws:ses:"
                + self.region
                + ":"
                + self.account
                + ":identity/"
                + domain_name,
                "SUPPORT_CHAT_URL": "https://discord.gg/nQVpNGGkYu",
                "APPLICATION_ADMIN": "consoleme_admin",
                "ACCOUNT_NUMBER": self.account,
                "ISSUER": domain_name,
                "SPOKE_ACCOUNTS": ",".join(spoke_accounts),
                "CONFIG_SECRET_NAME": CONFIG_SECRET_NAME,
            },
        )

        create_configuration_resource_provider = cr.Provider(
            self,
            "CreateConfigurationFileProvider",
            on_event_handler=create_configuration_lambda,
            log_retention=logs.RetentionDays.ONE_WEEK,
        )

        create_configuration_lambda_resource = cdk.CustomResource(
            self,
            "CreateConfigurationFile",
            service_token=create_configuration_resource_provider.service_token,
            removal_policy=cdk.RemovalPolicy.DESTROY,
            properties={"UUID": str(uuid4())},
        )

        create_configuration_lambda_resource.node.add_dependency(config_secret)



import base64
import json
import sys

import jwt
import requests
import tornado.httpclient
from jwt.algorithms import ECAlgorithm, RSAAlgorithm
from jwt.exceptions import (
    ExpiredSignatureError,
    ImmatureSignatureError,
    InvalidAudienceError,
    InvalidIssuedAtError,
    InvalidIssuerError,
)
from okta_jwt.utils import verify_exp, verify_iat

from consoleme.config import config
from consoleme.exceptions.exceptions import (
    MissingConfigurationValue,
    UnableToAuthenticate,
)

log = config.get_logger()


async def populate_oidc_config():
    http_client = tornado.httpclient.AsyncHTTPClient()
    metadata_url = config.get(
        "get_user_by_aws_alb_auth_settings.access_token_validation.metadata_url"
    )

    if metadata_url:
        res = await http_client.fetch(
            metadata_url,
            method="GET",
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "Accept": "application/json",
            },
        )
        oidc_config = json.loads(res.body)
    else:
        jwks_uri = config.get(
            "get_user_by_aws_alb_auth_settings.access_token_validation.jwks_uri"
        )
        if not jwks_uri:
            raise MissingConfigurationValue("Missing OIDC Configuration.")
        oidc_config = {
            "jwks_uri": jwks_uri,
        }

    # Fetch jwks_uri for jwt validation
    res = await http_client.fetch(
        oidc_config["jwks_uri"],
        method="GET",
        headers={
            "Content-Type": "application/x-www-form-urlencoded",
            "Accept": "application/json",
        },
    )
    oidc_config["jwks_data"] = json.loads(res.body)
    oidc_config["jwt_keys"] = {}
    for k in oidc_config["jwks_data"]["keys"]:
        key_type = k["kty"]
        key_id = k["kid"]
        if key_type == "RSA":
            oidc_config["jwt_keys"][key_id] = RSAAlgorithm.from_jwk(json.dumps(k))
        elif key_type == "EC":
            oidc_config["jwt_keys"][key_id] = ECAlgorithm.from_jwk(json.dumps(k))
    oidc_config["aud"] = config.get(
        "get_user_by_aws_alb_auth_settings.access_token_validation.client_id"
    )
    return oidc_config


async def authenticate_user_by_alb_auth(request):
    aws_alb_auth_header_name = config.get(
        "get_user_by_aws_alb_auth_settings.aws_alb_auth_header_name", "X-Amzn-Oidc-Data"
    )
    aws_alb_claims_header_name = config.get(
        "get_user_by_aws_alb_auth_settings.aws_alb_claims_header_name",
        "X-Amzn-Oidc-Accesstoken",
    )
    function = f"{__name__}.{sys._getframe().f_code.co_name}"
    log_data = {"function": function}
    encoded_auth_jwt = request.request.headers.get(aws_alb_auth_header_name)
    access_token = request.request.headers.get(aws_alb_claims_header_name)
    if not encoded_auth_jwt:
        raise Exception(f"Missing header: {aws_alb_auth_header_name}")
    if not access_token:
        raise Exception(f"Missing header: {aws_alb_claims_header_name}")

    jwt_headers = encoded_auth_jwt.split(".")[0]
    decoded_jwt_headers = base64.b64decode(jwt_headers)
    decoded_jwt_headers = decoded_jwt_headers.decode("utf-8")
    decoded_json = json.loads(decoded_jwt_headers)
    kid = decoded_json["kid"]
    # Step 2: Get the public key from regional endpoint
    url = "https://public-keys.auth.elb." + config.region + ".amazonaws.com/" + kid
    req = requests.get(url)
    pub_key = req.text
    # Step 3: Get the payload
    payload = jwt.decode(encoded_auth_jwt, pub_key, algorithms=["ES256"])
    email = payload.get(
        config.get("get_user_by_aws_alb_auth_settings.jwt_email_key", "email")
    )

    if not email:
        raise UnableToAuthenticate("Unable to determine user from ID Token")

    # Step 4: Parse the Access Token
    # User has already passed ALB auth and successfully authenticated
    access_token_pub_key = None
    jwt_verify = config.get("get_user_by_aws_alb_auth_settings.jwt_verify", True)
    access_token_verify_options = {"verify_signature": jwt_verify}
    oidc_config = {}
    algorithm = None
    try:
        if jwt_verify:
            oidc_config = await populate_oidc_config()
            header = jwt.get_unverified_header(access_token)
            key_id = header["kid"]
            algorithm = header["alg"]
            if algorithm == "none" or not algorithm:
                raise UnableToAuthenticate(
                    "Access Token header does not specify a signing algorithm."
                )
            access_token_pub_key = oidc_config["jwt_keys"][key_id]

        decoded_access_token = jwt.decode(
            access_token,
            access_token_pub_key,
            algorithms=[algorithm],
            options=access_token_verify_options,
            audience=oidc_config.get("aud"),
            issuer=oidc_config.get("issuer"),
        )
        # Step 5: Verify the access token.
        if not jwt_verify:
            verify_exp(access_token)
            verify_iat(access_token)

        # Extract groups from tokens, checking both because IdPs aren't consistent here
        for token in [decoded_access_token, payload]:
            groups = token.get(
                config.get("get_user_by_aws_alb_auth_settings.jwt_groups_key", "groups")
            )
            if groups:
                break

    except jwt.exceptions.DecodeError as e:
        # This exception occurs when the access token is not JWT-parsable. It is expected with some IdPs.
        log.debug(
            {
                **log_data,
                "message": (
                    "Unable to decode claims token. This is expected for some Identity Providers."
                ),
                "error": e,
                "user": email,
            }
        )
        log.debug(log_data, exc_info=True)
        groups = []
    except (
        ExpiredSignatureError,
        ImmatureSignatureError,
        InvalidAudienceError,
        InvalidIssuedAtError,
        InvalidIssuerError,
    ) as e:
        # JWT Validation failed, log an error and revoke the ALB auth cookie
        log.debug(
            {
                **log_data,
                "message": (str(e)),
                "error": e,
                "user": email,
            }
        )
        log.debug(log_data, exc_info=True)
        request.clear_cookie("AWSELBAuthSessionCookie-0")
        request.redirect(request.request.uri)

        groups = []

    return {"user": email, "groups": groups}



import json  # I don't think I can use ujson here
from datetime import datetime
from decimal import Decimal

from deepdiff.model import PrettyOrderedSet


class SetEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (frozenset, set, PrettyOrderedSet)):
            return list(obj)
        if isinstance(obj, Decimal):
            return float(obj)
        if isinstance(obj, datetime):
            return obj.timestamp()
        if isinstance(obj, Exception):
            return str(obj)
        return json.JSONEncoder.default(self, obj)



import os

import ed25519

from consoleme.config import config
from consoleme.lib.plugins import get_plugin_by_name

log = config.get_logger("consoleme")
stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()


class Crypto:
    def __init__(self) -> None:
        self.load_secrets()

    def load_secrets(self) -> None:
        if not config.get("ed25519.signing_key"):
            # Generating keys on demand. This is useful for unit tests
            self.signing_key, self.verifying_key = ed25519.create_keypair(
                entropy=os.urandom
            )
            return

        signing_key_file = os.path.expanduser(config.get("ed25519.signing_key"))
        try:
            with open(signing_key_file, "rb") as signing_file:
                signing_key_str: str = signing_file.read()
        except FileNotFoundError:
            msg = "Unable to load signing key"
            log.error(msg, exc_info=True)
            raise Exception(msg)
        self.signing_key = ed25519.SigningKey(signing_key_str)
        verifying_key_file = config.get("ed25519.verifying_key")
        try:
            verifying_key_str = open(verifying_key_file, "rb").read()
        except FileNotFoundError:
            msg = "Unable to load verifying key"
            log.error(msg, exc_info=True)
            raise Exception(msg)
        self.verifying_key = ed25519.VerifyingKey(verifying_key_str)

    def sign(self, s: str) -> bytes:
        return self.signing_key.sign(str.encode(s), encoding="base64")

    def verify(self, s, sig):
        try:
            if not s:
                return False
            self.verifying_key.verify(sig, str.encode(s), encoding="base64")
            return True
        except ed25519.BadSignatureError:
            stats.count("verify.bad_sig")
            log.error("Bad signature", exc_info=True)
            return False



from typing import Any, Dict, List

from consoleme.config import config
from consoleme.handlers.base import BaseAPIV2Handler
from consoleme.lib.auth import can_admin_policies
from consoleme.lib.defaults import (
    PERMISSION_TEMPLATE_DEFAULTS,
    SELF_SERVICE_IAM_DEFAULTS,
)


class SelfServiceConfigHandler(BaseAPIV2Handler):
    allowed_methods = ["GET"]

    async def get(self):
        admin_bypass_approval_enabled: bool = can_admin_policies(self.user, self.groups)
        export_to_terraform_enabled: bool = config.get(
            "export_to_terraform_enabled", False
        )
        self_service_iam_config: dict = config.get(
            "self_service_iam", SELF_SERVICE_IAM_DEFAULTS
        )

        # Help message can be configured with Markdown for link handling
        help_message: str = config.get("self_service_iam_help_message")

        self.write(
            {
                "admin_bypass_approval_enabled": admin_bypass_approval_enabled,
                "export_to_terraform_enabled": export_to_terraform_enabled,
                "help_message": help_message,
                **self_service_iam_config,
            }
        )


class PermissionTemplatesHandler(BaseAPIV2Handler):
    allowed_methods = ["GET"]

    async def get(self):
        """
        Returns permission templates.

        Combines permission templates from dynamic configuration to the ones discovered in static configuration, with a
        priority to the templates defined in dynamic configuration.

        If no permission_templates are defined in static configuration, this function will substitute the static
        configuration templates with PERMISSION_TEMPLATE_DEFAULTS.
        """
        permission_templates_dynamic_config: List[Dict[str, Any]] = config.get(
            "dynamic_config.permission_templates", []
        )

        permission_templates_config: List[Dict[str, Any]] = config.get(
            "permission_templates", PERMISSION_TEMPLATE_DEFAULTS
        )

        seen = set()
        compiled_permission_templates = []
        for item in [
            *permission_templates_dynamic_config,
            *permission_templates_config,
        ]:
            if item["key"] in seen:
                continue
            compiled_permission_templates.append(item)
            seen.add(item["key"])

        self.write({"permission_templates": compiled_permission_templates})
